# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from dataclasses import dataclass, asdict
import json
import re
from typing import Union, List, Dict, TYPE_CHECKING
from collections import defaultdict

from chipscopy.api import CoreType
from chipscopy.api._detail.ltx import Ltx, LtxCore, LtxProbe
from chipscopy.api._detail.debug_core import DebugCore

if TYPE_CHECKING:
    from chipscopy.client.axis_vio_core_client import AxisVIOCoreClient


@dataclass
class VIOProbe:
    """VIOProbe represents a single VIO probe connected to a VIO port.
    A VIO Port may be connected to many probes. Each probe is a slice
    of the port as shown below.

    ::

                            --------
          ----------------- VIO Port ------------------------
                            --------
        msb                                                lsb
        N-1                        47          16 15        0
         -----------------------------------------------------
        | boo[]   | ... | baz[]   | bar[31:0]    | foo[15:0] |
        -----------------------------------------------------
                                   ^            ^
                                   |          port_bit_
                                   |          offset
                                   |            16
                                   |            |
                                 bus_left      bus_right
                                  _index       _index
                                  31           0

    Normally the left index is the most significant probe index, and the right
    index is the least significant. They match the HDL index of what was probed
    in the design.

    """

    probe_name: str
    """Represents the HDL net or bus name. It does not contain the brackets
    of the bus name [x:y]."""
    direction: str
    """in or out. IN is a monitor connected to something in the design. Out is
    a virtual button the VIO drives."""
    is_bus: bool
    """True if this probe is a vector, False if this probe is a scalar."""
    bus_left_index: int
    """``probe_name[bus_left_index::bus_right_index]`` - When **is_bus == True**, this is the left index of the probe.
    When **is_bus == False**, this unused."""
    bus_right_index: int
    """``probe_name[bus_left_index::bus_right_index]`` - When **is_bus == True**, this is the right index of the probe.
    When **is_bus == False**, this unused."""
    port_index: int
    """VIO port number of the input or output port."""
    port_bit_offset: int
    """Bit offset into the probe port that indicates where the probe data begins in the port."""

    def __init__(self, ltx_probe: LtxProbe):
        """Build a Probe based on the ltx probe information"""
        self.probe_name = ltx_probe.name
        self.direction = ltx_probe.direction.lower()
        self.is_bus = ltx_probe.is_bus
        self.bus_left_index = ltx_probe.bus_left_index
        self.bus_right_index = ltx_probe.bus_right_index
        self.port_index = ltx_probe.port_index
        self.port_bit_offset = 0

    def __str__(self):
        return self.probe_name

    def __repr__(self):
        return json.dumps(dataclasses.asdict(self), indent=4)

    @property
    def port_name(self):
        """Represents the VIO port name in the IP (like probe_in0, probe_out1, etc)"""
        return f"probe_{self.direction}{self.port_index}"


class VIO(DebugCore["AxisVIOCoreClient"]):
    """This class contains the main API to use the VIO (Virtual Input/Output)
    debug core. VIO monitors elements of a running design in hardware with
    probe inputs and drives elements in the design with probe outputs.

    This API has methods for querying and controlling VIO ports directly.
    This is useful for lower level control, or direct VIO control when no LTX
    file is available.

    If an LTX file is available, the higher level read_probes and write_probes
    methods are available to debug at the higher level HDL context. HDL nets and
    bus names from Vivado are automatically converted to the correct VIO probe
    port when reading or writing.
    """

    def __init__(self, vio_tcf_node, *, ltx: Ltx = None):
        super(VIO, self).__init__(core_type=CoreType.AXIS_VIO, core_tcf_node=vio_tcf_node)

        self.name = vio_tcf_node.props["Name"]
        self.instance_name = None
        self.ltx_core = None
        self.uuid = self.core_info.uuid
        (
            self.port_in_widths,
            self.port_out_widths,
            self.has_activity,
        ) = self._read_port_info_from_hardware()
        self._setup_vio_from_ltx(ltx)

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "uuid": self.uuid, "instance_name": self.instance_name}

    def __repr__(self):
        input_ports = {}
        for idx, val in enumerate(self.port_in_widths):
            input_ports[f"probe_in{idx}"] = val
        output_ports = {}
        for idx, val in enumerate(self.port_out_widths):
            output_ports[f"probe_out{idx}"] = val
        probes = [asdict(probe) for probe in self.probes]
        ret_dict = {
            "name": self.name,
            "type": str(self.core_type),
            "uuid": self.uuid,
            "instance": self.instance_name,
            "input_ports": input_ports,
            "output_ports": output_ports,
            "probes": probes,
        }
        json_dict = json.dumps(ret_dict, indent=4)
        return json_dict

    def __str__(self):
        return self.name

    ###########################################################################
    # VIO CORE AND PORT CONTROL
    ###########################################################################
    @property
    def is_activity_supported(self) -> bool:
        """True if this vio core supports activity detection"""
        return self.has_activity

    @property
    def port_in_count(self) -> int:
        """Number of VIO input ports in the debug core. Input ports monitor
        activity in the design.
        """
        return len(self.port_in_widths)

    @property
    def port_out_count(self) -> int:
        """Number of VIO output ports in debug core. Output ports are virtual
        buttons.
        """
        return len(self.port_out_widths)

    @property
    def port_names(self) -> List[str]:
        """List of all VIO port names (input and output) for this VIO core.

        VIO cores name ports as ``probe_in#`` for input ports, and
        ``probe_out#`` for output ports. '#' represents the port index and
        begins at 0.
        """
        port_names = []
        for idx in range(self.port_in_count):
            port_names.append(f"probe_in{idx}")
        for idx in range(self.port_out_count):
            port_names.append(f"probe_out{idx}")
        return port_names

    def reset_vio(self):
        """
        Resets the VIO core by resetting all probe outputs to their initial
        values.
        """
        self.core_tcf_node.reset_core()

    def read_ports(self, port_names: Union[str, List[str]] = None) -> Dict[str, Dict]:
        """Read VIO port values from hardware. Gets the current values from
        hardware for selected input_ports and output_ports.

        Args:
            port_names: **(optional)** List of port names (see
                ``vio_port_names``). The default is all ports.

        Returns:
            dict:
                dict: (port_name, {'value': value, 'activity': activity}).

                'value' is always returned as an integer for each port.

                'activity' is only included for vio input ports when the
                activity detection is turned on. Activity does not exist
                for vio output ports. Activity is returned as one character per
                bit of the probe. 'R', 'F', 'N', 'B' represent edges -  Rising,
                Falling, None, Both.

                port_data returned as a dictionary in the following format:

                ::

                    port_data = {
                                   probe_in0: {'value': 3, 'activity': 'RR'},
                                   ... ,
                                   probe_out0: {'value': 456},
                                   ...
                                }

        """
        # TODO: Document the order of 'activity' bits in the return string
        if isinstance(port_names, str):
            port_names = [port_names]
        if port_names is None:
            port_names = self.port_names
        port_in_data = self.core_tcf_node.refresh_port_in_data()
        port_out_data = self.core_tcf_node.refresh_port_out_data()
        port_data = {}
        for port_name in port_names:
            port_name_split = re.split("^probe_(in|out)(\\d+)$", port_name)
            assert len(port_name_split) == 4
            _, port_direction, port_index_str, _ = port_name_split  # ['', 'in|out', idx, '']
            port_index = int(port_index_str)
            if port_direction == "in":
                assert port_index < self.port_in_count
                val = port_in_data["In"][port_index]
                if "Activity" in port_in_data:
                    activity = port_in_data["Activity"][port_index]
                    port_data[port_name] = {"value": val, "activity": activity}
            elif port_direction == "out":
                assert port_index < self.port_out_count
                val = port_out_data["Out"][port_index]
                port_data[port_name] = {"value": val}
            else:
                raise ValueError(f"Invalid port name - {port_name}")
        return port_data

    def write_ports(self, port_values: Dict[str, int]):
        """Write values to VIO port outputs in hardware. Port names follow
        the VIO convention port in and port out naming.
        See ``vio_port_names``.

        Args:
            port_values: dict. key=port_name, value=port_value.
                port_name is a VIO output port. port_value is a python int
                value.

                ::

                      write_ports({'probe_out0': 1000,
                                   'probe_out1': 0x100})

        Returns:
            Nothing
        """
        assert isinstance(port_values, dict)
        port_data = {}
        for port_name, port_value in port_values.items():
            port_name_split = re.split("^probe_out(\\d+)$", port_name)
            if len(port_name_split) != 3:
                raise KeyError(f"Invalid port name {port_name}")
            _, port_index_str, _ = port_name_split  # ['', idx, '']
            port_index = int(port_index_str)
            port_data[port_index] = port_value
        assert len(port_data) > 0
        self.core_tcf_node.commit_port_out_data(port_data)

    ###########################################################################
    # VIO PROBE ACCESS
    ###########################################################################
    @property
    def probes(self) -> List[VIOProbe]:
        """
        Get probes attached to the VIO core and associated port information.

        Returns:
            List of VIOProbes
        """
        probe_list = []
        if self.ltx_core:
            ltx_probe_dict = self._get_probe_to_port_map(self.ltx_core.probes)
            for port_name, ltx_probe_list in ltx_probe_dict.items():
                for ltx_probe in ltx_probe_list:
                    probe_list.append(VIOProbe(ltx_probe))
        return probe_list

    @property
    def probe_names(self) -> List[str]:
        """
        Returns:
             List of probe names from the LTX file
        """
        probe_names = []
        if self.ltx_core:
            for probe in self.ltx_core.probes:
                probe_names.append(probe.name)
        return probe_names

    def clear_activity(self):
        """Clear any stale activity indicator data. This really just does a
        dummy read of all probes.

        Returns:
            Nothing
        """
        self.read_probes()

    def read_probe_activity(self, probe_names: List[str] or str = None) -> Dict[str, str]:
        """Read probe activity values from hardware.  Reading activity has the
        side effect of clearing activity registers in the VIO IP in hardware.

        Activity tracking is only available on input ports, and must be enabled
        when generating the VIO IP.

        Activity is returned as a string of characters representing activity
        for each bit.

        ::

            'R' = rising edge detected
            'F' = falling edge detected
            'B' = both edges detected
            'N' = no edge detected
            'X' = activity not supported on port (ex: vio output port)

        Args:
            probe_names: Optional list of probe names (see ``vio_probe_names``). Empty returns all probe activity.

        Returns:
            dict:
                dict: {probe_name: activity_string, ...}.
        """
        probe_dict = self.read_probes(probe_names)
        retval = {}
        for k, v in probe_dict.items():
            retval[k] = v["activity"]
        return retval

    def read_probe_values(self, probe_names: List[str] or str = None) -> Dict[str, int]:
        """Read probe values from hardware. Reading probes has a side effect of
        clearing activity. If you need both values and activity, call read_probes().

        Args:
            probe_names: Optional list of probe names (see ``vio_probe_names``). Empty returns all probe values.

        Returns:
            dict:
                dict: {probe_name:  value, ...}.
                'value' is always returned as an integer for each port.
        """
        probe_dict = self.read_probes(probe_names)
        retval = {}
        for k, v in probe_dict.items():
            retval[k] = v["value"]
        return retval

    def read_probes(self, probe_names: List[str] or str = None) -> Dict[str, Dict]:
        """Read probe values from hardware. Gets the current integer and activity values from
        hardware for selected probes. Output probes to not support activity and will return
        'X' for all activity values.

        Args:
            probe_names: Optional list of probe names (see ``vio_probe_names``). Empty returns all probe values.

        Returns:
            dict:
                dict: (probe_name, {'value': value, 'activity': activity}).

                'value' is always returned as an integer for each port.

                'activity' is only included for vio input probes when the
                activity detection is turned on. Activity does not exist
                for vio output probes. Activity is returned as one character per
                bit of the probe. 'R', 'F', 'N', 'B' represent edges -  Rising,
                Falling, None, Both.
                'X' returned for activity indicates it is not supported on that probe.

                ::

                    port_data = {
                                   foo_probe: {'value': 3, 'activity': 'RR'},
                                   ... ,
                                   bar_probe: {'value': 456},
                                   ...
                                }
        """
        self.core_tcf_node.refresh_probe()
        report_results: Dict = self.core_tcf_node.report_probe()
        retval: Dict = {}
        if isinstance(probe_names, str):
            probe_names = [probe_names]
        elif probe_names is None:
            # None defaults to all probes
            probe_names = report_results.keys()
        for probe_name in probe_names:
            probe_value = report_results[probe_name]["value"]
            probe_activity = report_results[probe_name]["activity"]
            retval[probe_name] = {"value": probe_value, "activity": probe_activity}
        return retval

    def write_probes(self, probe_values: Dict[str, int] = None):
        """Write values to VIO probe outputs in hardware. See
        ``vio_probe_names``.

        Args:
            probe_values: dict. key=probe_name, value=probe_value.
                probe_name is a VIO output probe. probe_value is a python int
                value.

                ::

                      write_probes({'foo': 1000,
                                    'bar': 0x100})

        Returns:
            Nothing
        """
        assert isinstance(probe_values, dict)
        self.core_tcf_node.commit_probe(probe_values)

    def _read_port_info_from_hardware(self):
        ports_info_dict = self.core_tcf_node.get_ports_info()
        port_in_widths: [int] = []
        port_in_count = ports_info_dict["port_in_count"]
        if port_in_count > 0:
            port_in_widths = ports_info_dict["port_in_widths"]
        port_out_widths: [int] = []
        port_out_count = ports_info_dict["port_out_count"]
        if port_out_count > 0:
            port_out_widths = ports_info_dict["port_out_widths"]
        has_activity: bool = ports_info_dict["has_activity_detector"]
        return port_in_widths, port_out_widths, has_activity

    @staticmethod
    def _get_probe_to_port_map(ltx_probes: [LtxProbe]) -> Dict[str, List[LtxProbe]]:
        # Build the probe to port mapping. Associates the logical
        # LTX port data to the physical VIO hardware port configuration.
        # Group probes by port index and order probes from smallest to largest
        # offset { port_index : [probe, ..., probe] ... }
        # There may be multiple logical probes for each physical port.

        probes_by_port = defaultdict(list)
        for ltx_probe in ltx_probes:
            # TODO: Need to handle probe slices... That means sort by bit index
            #       and handle multiple probes for a single port
            direction = ltx_probe.direction.lower()
            port_key = f"probe_{direction}{ltx_probe.port_index}"
            probes_by_port[port_key].append(ltx_probe)
        return probes_by_port

    def _create_probes(self, probe_to_port_map: Dict[str, List[LtxProbe]]):
        for port_index in sorted(probe_to_port_map.keys()):
            probes_for_port = probe_to_port_map[port_index]
            for ltx_probe in probes_for_port:
                probe = VIOProbe(ltx_probe)
                # print("_create_probes - CREATED PROBE: ", probe)
                port_name = f"{probe.direction}{probe.port_index}"
                probe_to_create = {
                    "name": probe.probe_name,
                    "net": probe.probe_name,
                    "map": port_name,
                }
                self.core_tcf_node.define_probe([probe_to_create])

    def _setup_vio_from_ltx(self, ltx: Ltx):
        # Clear out any old LTX leftovers
        self.instance_name = None
        self.core_tcf_node.undefine_probe(probe_name="All")
        self.ltx_core = None
        if ltx:
            ltx_core: LtxCore = ltx.get_core(core_type=self.core_type, uuid=self.uuid)
            self.ltx_core = ltx_core
            assert ltx_core is not None
            self.instance_name = ltx_core.cell_name
            self.name = ltx_core.cell_name
            probe_to_port_map = self._get_probe_to_port_map(ltx_core.probes)
            self._create_probes(probe_to_port_map)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, ClassVar, Dict
from dataclasses import dataclass

from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.tcf.services import DoneHWCommand


@dataclass
class Sysmon(DebugCore["SysMonCoreClient"]):
    """Sysmon is the top level API for accessing System Monitor in Versal."""

    def __init__(self, sysmon_tcf_node):
        super(Sysmon, self).__init__(CoreType.SYSMON, sysmon_tcf_node)

        self.name = sysmon_tcf_node.props["Name"]

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name}

    def __str__(self):
        return self.name

    def initialize_sensors(self, done: DoneHWCommand = None):
        """
        Initializes all of the server side constructs for sysmon, required for most subsequent APIs.

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            No data is returned directly from this method.

        """
        self.core_tcf_node.initialize_sensors(done)

    def refresh_measurement_schedule(self, done: DoneHWCommand = None) -> Dict[int, str]:
        """
        Dynamically scans the SysMon to determine the sensors scheduled into the system, in order.


        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            Returns a dict where the keys are the root IDs and the values are the named sensors.

        """
        sensors = self.core_tcf_node.refresh_measurement_schedule(done)
        return sensors

    def get_measurements(self, done: DoneHWCommand = None) -> List[object]:
        """
        Advanced API

        Returns the measurement objects

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            Returns a list of measurement configuration objects

        """
        sensors = self.core_tcf_node.get_measurements(done)
        return sensors

    def get_all_sensors(self, done: DoneHWCommand = None) -> dict:
        """
        This function is identical to ``get_supported_sensors`` in customer releases. For VnC teams this API offers
        internal sensors. For information on how to use this API please contact @dkopelov or #chipscope.

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            A list of strings representing all sensors for the connected device.

        """
        sensors = self.core_tcf_node.get_all_sensors(done)
        return sensors

    def configure_measurement_schedule(
        self, measurements: Dict[str, str], done: DoneHWCommand = None
    ) -> None:
        self.core_tcf_node.configure_measurement_schedule(measurements, done)

    def stream_sensor_data(self, interval: int, done: DoneHWCommand = None) -> None:
        """
        This API configures polling in hw_server to stream the data from the measurement schedule (including device
        temps) at the specified ``interval``.

        To process the incoming data, a `dm.NodeListener` must be declared and registered on this node.


        node_listener = dm.NodeListener(sysmon)

        session.chipscope_view.add_node_listener(node_listener)

        Args:
            interval: poll period (in ms)
            done: Optional command callback that will be invoked when the response is received.
        """
        self.core_tcf_node.stream_sensor_data(interval, done)



# *****************************************************************************
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import binascii
import json

# channel states
STATE_OPENING = 0
STATE_OPEN = 1
STATE_CLOSED = 2


class TraceListener(object):
    def onMessageReceived(self, t, token, service, name, data):
        pass

    def onMessageSent(self, t, token, service, name, data):
        pass

    def onChannelClosed(self, error):
        pass


class Proxy(object):
    def onCommand(self, token, service, name, data):
        pass

    def onEvent(self, service, name, data):
        pass

    def onChannelClosed(self, error):
        pass

_token_cnt = 0


class Token(object):
    def __init__(self, tokenID=None, listener=None):
        if tokenID is None:
            global _token_cnt
            tokenID = str(_token_cnt)
            _token_cnt += 1
        else:
            if isinstance(tokenID, bytearray):
                tokenID = tokenID.decode('utf-8')
            else:
                tokenID = str(tokenID)
        self.id = tokenID
        self.listener = listener

    def getID(self):
        return self.id

    def getListener(self):
        return self.listener

    def cancel(self):
        return False

    def __str__(self):
        return self.id


class ChannelListener(object):
    """
    Channel listener interface.
    """

    def onChannelOpened(self):
        """
        Called when a channel is opened or redirected.
        """
        pass

    def onChannelClosed(self, error):
        """
        Called when channel closed. If it is closed because of an error,
        'error' parameter will describe the error. 'error' is None if channel
        is closed normally by calling Channel.close().
        @param error - channel exception or None
        """
        pass

    def congestionLevel(self, level):
        """
        Notifies listeners about channel out-bound traffic congestion level
        changes.
        When level > 0 client should delay sending more messages.
        @param level - current congestion level
        """
        pass


class EventListener(object):
    """
    A generic interface for service event listener.
    Services usually define a service specific event listener interface,
    which is implemented using this generic listener.
    Clients should user service specific listener interface,
    unless no such interface is defined.
    """
    svc_name = "<unknown>"

    def event(self, name, data):
        """
        Called when service event message is received
        @param name - event name
        @param data - event arguments encoded as bytearray
        """
        pass


class CommandServer(object):
    """
    Command server interface.
    This interface is to be implemented by service providers.
    """
    def command(self, token, name, data):
        """
        Called every time a command is received from remote peer.
        @param token - command handle
        @param name - command name
        @param data - command arguments encoded into array of bytes
        """
        pass


class CommandListener(object):
    """
    Command listener interface. Clients implement this interface to
    receive command results.
    """
    def progress(self, token, data):
        """
        Called when progress message (intermediate result) is received
        from remote peer.
        @param token - command handle
        @param data - progress message arguments encoded into array of bytes
        """
        pass

    def result(self, token, data):
        """
        Called when command result received from remote peer.
        @param token - command handle
        @param data - command result message arguments encoded into array of
        bytes
        """
        pass

    def terminated(self, token, error):
        """
        Called when command is terminated because communication channel was
        closed or command is not recognized by remote peer.
        @param token - command handle
        @param error - exception that forced the channel to close
        """
        pass


def queue_bytes(arg):
    return b'(' + str(len(arg)).encode('utf-8') + b')' + arg


def toJSONSequence(args):
    if args is None:
        return None
    sequence = bytearray()
    for arg in args:
        if isinstance(arg, bytes):
            sequence += queue_bytes(arg)
        elif isinstance(arg, bytearray):
            sequence += queue_bytes(arg)
        elif hasattr(arg, '__json__'):
            sequence += arg.__json__()
        else:
            sequence += json.dumps(arg, separators=(',', ':'), cls=TCFJSONEncoder).encode('utf-8')
        sequence.append(0)
    return sequence


def parse_bin(s):
    n, _, data = s.partition(b')')
    n = int(n[1:].decode('utf-8'))
    return data[:n], data[n+1:]


def chomp_seq(s):
    if s and len(s) > 1 and s[-1] == 0:
        del s[-1]
    return s


def fromJSONSequence(s):
    objects = []
    #chomp_seq(s)
    while s and len(s) > 0:
        ch = s[0]
        if ch == b'('[0]:
            data, s = parse_bin(s)
            objects.append(data)
            continue
        j, _, s = s.partition(b'\x00')
        if j and len(j):
            j = chomp_seq(j)
            # NOTE - This decode and load will fail if an empty list is passed from Vivado
            #  This is a known issue. Solution - Don't send an empty list using the XHWPropertyMap
            objects.append(json.loads(j.decode('utf-8')))
        else:
            objects.append(None)
    return objects


def dumpJSONObject(obj):
    return json.dumps(obj, separators=(',', ':'), cls=TCFJSONEncoder)


def toByteArray(data):
    if data is None:
        return None
    t = type(data)
    if t is bytearray:
        return data
    else:
        return bytearray(binascii.a2b_base64(data))


class TCFJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if hasattr(o, '__json__'):
            return o.__json__()
        elif hasattr(o, '__iter__'):
            return tuple(o)
        else:
            return json.JSONEncoder.default(self, o)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import csv
from pathlib import Path
from typing import Dict, List, TYPE_CHECKING
from dataclasses import dataclass
from collections import OrderedDict

try:
    import pandas as pd
    import plotly.graph_objects as go

    _plotting_pkgs_available = True
except ImportError:
    _plotting_pkgs_available = False

try:
    from IPython import get_ipython
    from IPython.display import Image, display

    _jupyter_available = True
except ImportError:
    get_ipython = None
    Image = None
    display = None
    _jupyter_available = False

from chipscopy.dm import request
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.utils.printer import PercentProgressBar

if TYPE_CHECKING:
    from chipscopy.client.ddrmc_client import DDRMCClient


@dataclass
class DDR(DebugCore["DDRMCClient"]):
    """
    This class contains the top level API functions to interact with the
    integrated DDR memory controller debug core(s) on Versal devices.

    """

    def __init__(self, ddr_node):
        super(DDR, self).__init__(CoreType.DDR, ddr_node)
        self.node_name = ddr_node.props["Name"]
        self.mc_index = self.node_name[-1]
        self.name = "ddr_" + self.mc_index
        self.ddr_node = self.core_tcf_node
        self.is_enabled = self.is_user_enabled()
        self.eye_scan_data = []

    def get_cal_status(self):
        """
        Get the main calibration status of DDRMC

        Args:

        Returns:
            status in string

        """

        status = self.ddr_node.get_cal_status()

        return status

    def refresh_cal_status(self, done: DoneHWCommand = None):
        """
        Refresh the main calibration status of DDRMC

        Args:
            done: Optional command callback that will be invoked when the response is received

        Returns:

        """

        self.ddr_node.refresh_cal_status(done)

    def get_cal_stages(self):
        """
        Get the decoded calibration stages results happened from a
        calibration run

        Args:

        Returns:
            A dictionary where
                Key = calibration stage name
                Val = stage calibration result

        """
        results = {}
        results = self.ddr_node.get_property_group(["stages"])

        return results

    def refresh_cal_margin(self, done: DoneHWCommand = None):
        """
        Refresh the calibration windows margin analysis of DDRMC

        Args:
            done: Optional command callback that will be invoked when the response is received

        Returns:

        """

        self.ddr_node.refresh_cal_margin(done)

    def refresh_health_status(self, done: DoneHWCommand = None):
        """
        Refresh the overall health status analysis of DDRMC

        Args:
            done: Optional command callback that will be invoked when the response is received

        Returns:

        """

        self.ddr_node.refresh_health_status(done)

    def get_cal_margin_mode(self):
        """
        Get the availability of windows margin mode generated from a
        calibration run

        Args:

        Returns:
            A dictionary where
                Key = name of margin mode
                Val = True or False value indicating the availability

        """
        results = {}

        results = self.ddr_node.get_cal_margin_mode()

        return results

    def is_user_enabled(self):
        """
        Find out whether the DDRMC core is user enabled or not after
        Versal device configuration

        Args:

        Returns:
            Status value in Bool

        """
        # if not being powered by NoC clock, it is automatically disabled
        clock_off = self.ddr_node.is_clock_gated()

        if clock_off:
            return False
        else:
            result = {}
            key = "pcsr_control_ub_init"
            result = self.ddr_node.get_property(key)

            if not result[key]:
                return False
            else:
                return True

    def get_property(self, property_names=None, done: DoneHWCommand = None):
        """
        Get the property value mapped to DDRMC core

        Args:
            property_names: single string or list of string of property names
                (if None specified, all properties available will be queried and returned)

            done: Optional command callback that will be invoked when the response is received

        Returns:
            A dictionary, where
                Key = property name
                Val = property value

        """
        results = {}

        results = self.ddr_node.get_property(property_names, done)

        return results

    def __get_configuration(self) -> OrderedDict:
        configs = OrderedDict()
        is_lrdimm = False

        configs["DDRMC Core Name"] = self.name

        results = self.ddr_node.get_property("mem_type")
        val = results["mem_type"]
        if val == 1:
            configs["Memory Interface"] = "DDR4"
        elif val == 2:
            configs["Memory Interface"] = "LPDDR4"
        else:
            configs["Memory Interface"] = "Unknown"

        results = self.ddr_node.get_property("dimm_type")
        val = results["dimm_type"]
        if val == 0:
            configs["Device Type"] = "Component"
        elif val == 1:
            configs["Device Type"] = "UDIMM"
        elif val == 2:
            configs["Device Type"] = "RDIMM"
        elif val == 3:
            configs["Device Type"] = "LRDIMM"
            is_lrdimm = True
        else:
            configs["Device Type"] = "Unknown"

        results = self.ddr_node.get_property("slots")
        slots = results["slots"]
        results = self.ddr_node.get_property("phy_ranks")
        phy_ranks = results["phy_ranks"]
        results = self.ddr_node.get_property("mem_ranks")
        mem_ranks = results["mem_ranks"]

        if is_lrdimm:
            configs["Slots"] = str(phy_ranks)
            configs["Ranks"] = str(mem_ranks)
        else:
            configs["Slots"] = str(slots)
            configs["Ranks"] = str(phy_ranks)

        results = self.ddr_node.get_property("bytes")
        bytes = results["bytes"]
        results = self.ddr_node.get_property("nibbles")
        nibbles = results["nibbles"]
        results = self.ddr_node.get_property("bits_per_byte")
        bits = results["bits_per_byte"]
        data_width = bytes * bits
        configs["Data Width"] = str(data_width)
        configs["Bytes"] = str(bytes)
        configs["Nibbles"] = str(nibbles)
        configs["Bits per Byte"] = str(bits)

        freq_zero = "Unknown"
        freq_one = "Unknown"
        results = self.ddr_node.get_property("f0_period")
        period = results["f0_period"]
        if (period != "0") and (period != ""):
            freq0 = int(1000000 / float(period) + 0.5)
            freq_zero = str(freq0) + " MHz"
        results = self.ddr_node.get_property("f1_period")
        period = results["f1_period"]
        if (period != "0") and (period != ""):
            freq1 = int(1000000 / float(period) + 0.5)
            freq_one = str(freq1) + " MHz"
        results = self.ddr_node.get_property("dual_freq_en")
        dual_freq = results["dual_freq_en"]
        configs["Memory Frequency 0"] = freq_zero
        if int(dual_freq) > 0:
            configs["Memory Frequency 1"] = freq_one

        return configs

    def __get_cal_margins(self, mode_base: str, left_margs: Dict, right_margs: Dict, centers: Dict):
        group_name = mode_base + "_left"
        results = self.ddr_node.get_property_group([group_name])
        left_margs.clear()
        left_margs.update(results)
        group_name = mode_base + "_right"
        results = self.ddr_node.get_property_group([group_name])
        right_margs.clear()
        right_margs.update(results)
        group_name = mode_base + "_center"
        results = self.ddr_node.get_property_group([group_name])
        centers.clear()
        centers.update(results)

    def __output_read_margins(
        self,
        by_8_mode,
        bytes: int,
        nibbles: int,
        left_margs: Dict,
        right_margs: Dict,
        centers: Dict,
    ):
        left_marg_vals = sorted(left_margs.items())
        center_vals = sorted(centers.items())
        right_marg_vals = sorted(right_margs.items())

        if by_8_mode:
            for byte in range(bytes):
                left_marg = left_marg_vals[byte * 2][1]
                center = center_vals[byte * 2][1]
                right_marg = right_marg_vals[byte * 2][1]
                print(
                    "Byte ",
                    str(byte),
                    " Nibble 0  -",
                    "  Left Margin:  ",
                    str(left_marg[1]),
                    " (",
                    str(left_marg[0]),
                    ")  Center Point:  ",
                    str(center[1]),
                    " (",
                    str(center[0]),
                    ")  Right Margin:  ",
                    str(right_marg[1]),
                    " (",
                    str(right_marg[0]),
                    ")",
                )

                left_marg = left_marg_vals[byte * 2 + 1][1]
                center = center_vals[byte * 2 + 1][1]
                right_marg = right_marg_vals[byte * 2 + 1][1]
                print(
                    "Byte ",
                    str(byte),
                    " Nibble 1  -",
                    "  Left Margin:  ",
                    str(left_marg[1]),
                    " (",
                    str(left_marg[0]),
                    ")  Center Point:  ",
                    str(center[1]),
                    " (",
                    str(center[0]),
                    ")  Right Margin:  ",
                    str(right_marg[1]),
                    " (",
                    str(right_marg[0]),
                    ")",
                )
        else:
            for nibble in range(nibbles):
                left_marg = left_marg_vals[nibble][1]
                center = center_vals[nibble][1]
                right_marg = right_marg_vals[nibble][1]
                print(
                    "Nibble ",
                    str(nibble),
                    "  -" "  Left Margin:  ",
                    str(left_marg[1]),
                    " (",
                    str(left_marg[0]),
                    ")  Center Point:  ",
                    str(center[1]),
                    " (",
                    str(center[0]),
                    ")  Right Margin:  ",
                    str(right_marg[1]),
                    " (",
                    str(right_marg[0]),
                    ")",
                )

    def __output_write_margins(
        self, bytes: int, left_margs: Dict, right_margs: Dict, centers: Dict
    ):
        left_marg_vals = sorted(left_margs.items())
        center_vals = sorted(centers.items())
        right_marg_vals = sorted(right_margs.items())

        for byte in range(bytes):
            left_marg = left_marg_vals[byte][1]
            center = center_vals[byte][1]
            right_marg = right_marg_vals[byte][1]
            print(
                "Byte ",
                str(byte),
                "  -" "  Left Margin:  ",
                str(left_marg[1]),
                " (",
                str(left_marg[0]),
                ")  Center Point:  ",
                str(center[1]),
                " (",
                str(center[0]),
                ")  Right Margin:  ",
                str(right_marg[1]),
                " (",
                str(right_marg[0]),
                ")",
            )

    def report(self, to_file=False, file_name=None):
        """
        Run a report on the current statuses and analytical data of the DDRMC

        Args:
            to_file: specify True to have the report saved to a file

            file_name: the file name or full path for the report file to be saved

        Returns:

        """
        out_file = None
        orig_out = None

        if to_file:
            if not file_name:
                print("\nNOTE: Need to specify a file name for the report output.")
            else:
                out_file = open(file_name, "w")
                print("\nNOTE: The report is being generated and saved as:", file_name)
                orig_out = sys.stdout
                sys.stdout = out_file

        # DDRMC Status & Message
        print("-------------------\n")
        print(" DDRMC Status \n")
        print("-------------------\n")
        self.refresh_cal_status()
        self.refresh_health_status()
        print("Calibration Status:  ", self.get_cal_status(), "\n")
        results = self.ddr_node.get_property("health_status")
        print("Overall Health:  ", results["health_status"], "\n")
        results = self.ddr_node.get_property("cal_message")
        print("Message:  ", results["cal_message"], "\n")

        # Check if there is error condition to report
        results = self.ddr_node.get_property("cal_error_msg")
        if results["cal_error_msg"] != "None":
            print("Error:  ", results["cal_error_msg"], "\n")

        # DDRMC ISR Registers
        print("\n-------------------\n")
        print(" Status Registers\n")
        print("-------------------\n")

        print("DDRMC ISR Table\n")
        results = self.ddr_node.get_property_group(["ddrmc_isr_e", "ddrmc_isr_w"])
        for key, val in sorted(results.items()):
            print("  ", key, ":  ", str(val))

        print("\nUB ISR Table\n")
        results = self.ddr_node.get_property_group(["ub_isr_e", "ub_isr_w"])
        for key, val in sorted(results.items()):
            print("  ", key, ":  ", str(val))

        # Memory configuration info
        print("\n----------------------------------\n")
        print(" Memory Configuration \n")
        print("----------------------------------\n")
        configs = self.__get_configuration()
        for key, val in configs.items():
            print(key, ":  ", val)

        # Memory calibration stages info
        print("\n-----------------------------------\n")
        print(" Calibration Stages Information \n")
        print("-----------------------------------\n")
        # Do not do calibration stages if a system error is detected
        sys_error = False
        if self.get_cal_status() == "SYSTEM ERROR":
            sys_error = True
        if not sys_error:
            results = self.get_cal_stages()
            for key, val in sorted(results.items()):
                print(key, ":  ", val)
        else:
            print(
                "\nNote: DDRMC system error is detected. Calibration stages decoding will not be provided for ",
                self.name,
                "\n",
            )

        # Cal Margin Analysis
        print("\n---------------------------------------\n")
        print(" Calibration Window Margin Analysis \n")
        print("---------------------------------------\n")

        if not sys_error:
            cal_margin_modes = {}
            left_margins = {}
            right_margins = {}
            center_points = {}
            is_by_8 = True
            num_byte = int(configs["Bytes"])
            num_nibble = int(configs["Nibbles"])
            self.refresh_cal_margin()
            cal_margin_modes = self.get_cal_margin_mode()
            if int(configs["Bits per Byte"]) == 4:
                is_by_8 = False

            # Main loop to go through dual frequencies
            for freq in range(2):
                base = "Frequency " + str(freq)
                # Read Simple
                key = "f" + str(freq) + "_rd_simp"
                if cal_margin_modes[key]:
                    # Rising Edge
                    self.__get_cal_margins(
                        key + "_rise", left_margins, right_margins, center_points
                    )
                    print(
                        "\n",
                        base,
                        " - Read Margin - Simple Pattern - Rising Edge Clock in pS and (delay taps):\n",
                    )
                    self.__output_read_margins(
                        is_by_8, num_byte, num_nibble, left_margins, right_margins, center_points
                    )
                    # Falling Edge
                    self.__get_cal_margins(
                        key + "_fall", left_margins, right_margins, center_points
                    )
                    print(
                        "\n",
                        base,
                        " - Read Margin - Simple Pattern - Falling Edge Clock in pS and (delay taps):\n",
                    )
                    self.__output_read_margins(
                        is_by_8, num_byte, num_nibble, left_margins, right_margins, center_points
                    )
                # Read Complex
                key = "f" + str(freq) + "_rd_comp"
                if cal_margin_modes[key]:
                    # Rising Edge
                    self.__get_cal_margins(
                        key + "_rise", left_margins, right_margins, center_points
                    )
                    print(
                        "\n",
                        base,
                        " - Read Margin - Complex Pattern - Rising Edge Clock in pS and (delay taps):\n",
                    )
                    self.__output_read_margins(
                        is_by_8, num_byte, num_nibble, left_margins, right_margins, center_points
                    )
                    # Falling Edge
                    self.__get_cal_margins(
                        key + "_fall", left_margins, right_margins, center_points
                    )
                    print(
                        "\n",
                        base,
                        " - Read Margin - Complex Pattern - Falling Edge Clock in pS and (delay taps):\n",
                    )
                    self.__output_read_margins(
                        is_by_8, num_byte, num_nibble, left_margins, right_margins, center_points
                    )
                # Write Simple
                key = "f" + str(freq) + "_wr_simp"
                if cal_margin_modes[key]:
                    self.__get_cal_margins(key, left_margins, right_margins, center_points)
                    print(
                        "\n",
                        base,
                        " - Write Margin - Simple Pattern - Calibration Window in pS and (delay taps):\n",
                    )
                    self.__output_write_margins(
                        num_byte, left_margins, right_margins, center_points
                    )
                # Write Complex
                key = "f" + str(freq) + "_wr_comp"
                if cal_margin_modes[key]:
                    self.__get_cal_margins(key, left_margins, right_margins, center_points)
                    print(
                        "\n",
                        base,
                        " - Write Margin - Complex Pattern - Calibration Window in pS and (delay taps):\n",
                    )
                    self.__output_write_margins(
                        num_byte, left_margins, right_margins, center_points
                    )
        else:
            print(
                "\nNote: DDRMC system error is detected. Margin Analysis will not be provided for ",
                self.name,
                "\n",
            )

        if out_file:
            sys.stdout = orig_out
            out_file.close()

        return

    def set_eye_scan_read_mode(self):
        """
        Set 2D eye scan to read mode

        Args:

        Returns:

        """
        self.ddr_node.set_property({"mgchk_rw_mode": 0})
        self.ddr_node.commit_property_group([])

        return

    def set_eye_scan_write_mode(self):
        """
        Set 2D eye scan to write mode

        Args:

        Returns:

        """
        self.ddr_node.set_property({"mgchk_rw_mode": 1})
        self.ddr_node.commit_property_group([])

        return

    def set_eye_scan_simple_pattern(self):
        """
        Set 2D eye scan traffic pattern to simple mode

        Args:

        Returns:

        """
        self.ddr_node.set_property({"mgchk_pattern_mode": 0})
        self.ddr_node.commit_property_group([])

        return

    def set_eye_scan_complex_pattern(self):
        """
        Set 2D eye scan traffic pattern to complex mode

        Args:

        Returns:

        """
        self.ddr_node.set_property({"mgchk_pattern_mode": 1})
        self.ddr_node.commit_property_group([])

        return

    def set_eye_scan_rank(self, rank_num: int):
        """
        Set the target rank number on the DDR for 2D eye scan to run on.
        There will be 4 total possible rank selections. For single slot designs, there will be quad ranks.
        For dual slot designs, there will be dual ranks per slot.
        2D Margin Scan can only be performed on one rank at a time, and not in parallel.

        Args:

            rank_num:
                Rank number valid between 0 and 3 on the DDR for 2D Margin to be run on.

        Returns:

        """
        if isinstance(rank_num, str):
            print("ERROR: Please enter an integer value from 0 to 3 for Slot/Rank selection.")
            return
        elif rank_num > 3:
            print("ERROR: Cannot enter a rank_num value larger than 3.")
            return

        self.ddr_node.set_property({"mgchk_rank": rank_num})
        self.ddr_node.commit_property_group([])

        return

    def get_eye_scan_vref_percentage(self, vref: int) -> float:
        """
        A user helper function that takes in an encoding vref value intended for the margin scan,
        and converts it into a percentage value based on the current scan mode
        and memory configuration detected from the current hardware.

        Args:

            vref:
                For Read mode, valid integer range between 0 to 1023
                For Write mode, valid integer range between 0 to 50

        Returns:

            A translated percentage number in floating point

        """
        if isinstance(vref, str):
            print("ERROR: Please re-enter an integer value for Vref.")
            return
        elif vref > 1023:
            print("ERROR: Cannot enter VRef values larger than 1023")
            return

        base_value = 0
        incr_size = 0
        ddr4_range_one = 60
        ddr4_range_two = 45
        lp4_range_one = 10
        lp4_range_two = 20.4
        ddr4_incr_size = 0.65
        lp4_incr_size = 0.4
        read_incr_size = 0.0976563
        prop_list = []

        prop_list.append("mgchk_rw_mode")
        prop_list.append("mem_type")
        prop_list.append("mgchk_def_wr_vref_range")

        results = self.ddr_node.get_property(prop_list)
        rw_mode = results["mgchk_rw_mode"]
        mem_type = results["mem_type"]
        write_vref_range = results["mgchk_def_wr_vref_range"]

        if rw_mode:
            if vref > 50:
                print("ERROR: Cannot enter Vref values larger than 50 for Write Margin mode")
                return
            if mem_type == 1:
                if write_vref_range == 0:
                    base_value = ddr4_range_one
                else:
                    base_value = ddr4_range_two
                incr_size = ddr4_incr_size
            elif mem_type == 2:
                if write_vref_range == 0:
                    base_value = lp4_range_one
                else:
                    base_value = lp4_range_two
                incr_size = lp4_incr_size
        else:
            incr_size = read_incr_size

        offset_value = vref * incr_size
        output_vref = base_value + offset_value

        if not rw_mode:
            output_vref = "%.3f" % output_vref
        else:
            output_vref = "%.2f" % output_vref

        return float(output_vref)

    def set_eye_scan_vref_min(self, vref: int) -> float:
        """
        Set 2D eye scan desired minimum vref value to scan

        Args:

            vref:
                For Read mode, valid integer range between 0 to 1023
                For Write mode, valid integer range between 0 to 50

        Returns:

            A translated percentage number in floating point if
            the input VRef is set successfully

        """
        percent_vref = self.get_eye_scan_vref_percentage(vref)

        if percent_vref is not None:
            vref = int(vref)
            self.ddr_node.set_property({"es_vref_min": vref})
            return percent_vref

    def set_eye_scan_vref_max(self, vref: int) -> float:
        """
        Set 2D eye scan desired maximum vref value to scan

        Args:

            vref:
                For Read mode, valid integer range between 0 to 1023
                For Write mode, valid integer range between 0 to 50

        Returns:

            A translated percentage number in floating point if
            the input VRef is set successfully

        """
        percent_vref = self.get_eye_scan_vref_percentage(vref)

        if percent_vref is not None:
            vref = int(vref)
            self.ddr_node.set_property({"es_vref_max": vref})
            return percent_vref

    def set_eye_scan_vref_steps(self, steps: int):
        """
        Set 2D eye scan desired number of vref steps to scan vertically

        Args:

            steps:
                Valid integer range between 1 to 1024 for Read mode.
                Valid integer range between 1 to 51 for Write mode.

        Returns:

        """
        results = self.ddr_node.get_property(["mgchk_rw_mode"])
        rw_mode = results["mgchk_rw_mode"]

        if isinstance(steps, str):
            print("ERROR: Please re-enter an integer value for Vref steps.")
            return
        elif rw_mode and (steps > 51):
            print("ERROR: Cannot enter Vref step sizes larger than 51 for Write Margin mode")
            return
        elif steps > 1024:
            print("ERROR: Cannot enter Vref step sizes larger than 1024 for Read Margin mode")
            return

        steps = int(steps)
        self.ddr_node.set_property({"es_vref_steps": steps})

        return

    def __update_eye_scan_data(self):
        edges = ["left", "right"]
        clocks = ["nqtr", "pqtr"]
        clock_defs = {
            "nqtr": "fall_",
            "pqtr": "rise_",
        }
        data_list = []

        engine_data = self.ddr_node.get_eye_scan_data()
        ps_factors = self.ddr_node.get_margin_ps_factors()
        data_length = len(engine_data)

        # If initial vref scan went wrong after enabled, needless to process data further
        if not data_length > 1:
            return

        scan_mode = engine_data[0]["config.mode"]
        nibble_count = engine_data[0]["config.nibbles"]
        byte_count = engine_data[0]["config.bytes"]
        data_list.append(engine_data[0])

        if "Read" in scan_mode:
            old_key_base = "rdmargin_"
            new_key_base = "read_"
            for data in engine_data[1:]:
                for nibble in range(int(nibble_count)):
                    for clock in clocks:
                        for edge in edges:
                            old_key = (
                                old_key_base
                                + clock
                                + "_"
                                + edge
                                + "_nibble"
                                + "{:02d}".format(nibble)
                            )
                            ps_name = new_key_base + clock_defs[clock] + "{:02d}".format(nibble)
                            new_key = ps_name + "_" + edge
                            new_key_ps = new_key + "_ps"
                            ps_factor = ps_factors[ps_name]
                            tap_value = data[old_key]
                            ps_value = int(round(tap_value * ps_factor))
                            data[new_key] = tap_value
                            data[new_key_ps] = ps_value
                            del data[old_key]
                temp_data = {key: val for key, val in data.items() if "wrmargin" not in key}
                new_data = OrderedDict(sorted(temp_data.items()))
                data_list.append(new_data)
        else:
            old_key_base = "wrmargin_"
            new_key_base = "write_"
            for data in engine_data[1:]:
                for byte in range(int(byte_count)):
                    for edge in edges:
                        old_key = old_key_base + edge + "_byte" + str(byte)
                        ps_name = new_key_base + "{:02d}".format(byte)
                        new_key = ps_name + "_" + edge
                        new_key_ps = new_key + "_ps"
                        ps_factor = ps_factors[ps_name]
                        tap_value = data[old_key]
                        ps_value = int(round(tap_value * ps_factor))
                        data[new_key] = tap_value
                        data[new_key_ps] = ps_value
                        del data[old_key]
                temp_data = {key: val for key, val in data.items() if "rdmargin" not in key}
                new_data = OrderedDict(sorted(temp_data.items()))
                data_list.append(new_data)

        self.eye_scan_data = data_list
        self.ddr_node.set_property({"es_data_need_update": 0})

    def __draw_eye_scan_graph(
        self, margin_data: List, base_name: str, margin_title: str, def_vref: float
    ):
        vrefs = []
        left_margs = []
        right_margs = []
        left_ps = []
        right_ps = []
        x_min = 0
        x_max = 0

        for data in margin_data:
            vrefs.append(float(data["config.vrefp"]))
            key_name = base_name + "_left"
            left_margs.append(int(data[key_name]))
            key_name += "_ps"
            left_ps.append(int(data[key_name]))
            key_name = base_name + "_right"
            right_margs.append(int(data[key_name]))
            key_name += "_ps"
            right_ps.append(int(data[key_name]))

        x_min = max(left_ps)
        x_max = max(right_ps)

        data_defs = {
            "VRef": vrefs,
            "Left_Margin": left_margs,
            "Right_Margin": right_margs,
            "Left_PS": left_ps,
            "Right_PS": right_ps,
        }

        df = pd.DataFrame(data_defs)
        fig = go.Figure()

        for col in df.columns[1:2]:
            df_left = df[col]
            ps_left = df[df.columns[3]]
            data_left = list(zip(df_left, ps_left))
            fig.add_trace(
                go.Bar(
                    x=-ps_left.values,
                    y=df[df.columns[0]],
                    orientation="h",
                    name=col,
                    customdata=data_left,
                    marker_color="#75ad0a",
                    hovertemplate="Taps: %{customdata[0]} <br>pS: %{customdata[1]} <br>VRef: %{y}",
                )
            )

        for col in df.columns[2:3]:
            df_right = df[col]
            ps_right = df[df.columns[4]]
            data_right = list(zip(df_right, ps_right))
            fig.add_trace(
                go.Bar(
                    x=ps_right,
                    y=df[df.columns[0]],
                    orientation="h",
                    name=col,
                    customdata=data_right,
                    marker_color="#0073e6",
                    hovertemplate="Taps: %{customdata[0]} <br>pS: %{customdata[1]} <br>VRef: %{y}",
                )
            )

        fig.add_trace(
            go.Scatter(
                x=[-x_min, x_max],
                y=[def_vref, def_vref],
                mode="lines",
                line=dict(
                    color="#92b1a9",
                    width=4,
                    dash="dot",
                ),
                name="DDRMC VRef: " + str(def_vref),
                customdata=[x_min, x_max],
                hovertemplate="%{customdata}",
            )
        )

        fig.update_layout(
            barmode="relative",
            height=600,
            width=900,
            bargap=0.01,
            plot_bgcolor="#0f263e",
            title=margin_title,
            xaxis=dict(title="Delay in pS"),
            yaxis=dict(title="VRef Percentage"),
            legend_orientation="v",
            legend_x=1.0,
            legend_y=0.0,
        )

        return fig

    def save_eye_scan_data(self, file_name=None):
        """
        Save the eye scan 2D margin data from the most recent run to a file in CSV format

        Args:

            file_name: the file name or full path for the data file to be saved.
                       A default file name will be given if none is supplied.

        Returns:

        """
        out_file = None

        if not file_name:
            # Gives a default for users
            file_name = "eye_scan_data"

        file_name = os.path.splitext(file_name)[0] + ".csv"
        self.__update_eye_scan_data()

        if not self.eye_scan_data:
            print("ERROR: No scan data is found. Please re-run 2D eye scan.")
            return

        data_list = self.eye_scan_data
        keys = list(data_list[0].keys())
        more_keys = list(data_list[1].keys())
        keys.extend(more_keys)

        with open(file_name, "w", newline="") as out_file:
            data_writer = csv.DictWriter(out_file, keys)
            data_writer.writeheader()
            data_writer.writerows(data_list)
            print("\nInfo: Data are being formatted and saved to:", out_file.name)

    def load_eye_scan_data(self, file_name):
        """
        Load the eye scan 2D margin data from a data file

        Args:

            file_name: the file name or full path for the data file to be read and load

        Returns:

        """
        file_path = Path(file_name)
        data_list = []

        with file_path.open(mode="r") as data_file:
            reader = csv.DictReader(data_file)
            data_reader = list(reader)
            for data in data_reader:
                data_list.append(data)
            file_name = data_file.name

        self.eye_scan_data = data_list
        self.ddr_node.set_property({"es_data_need_update": 0})
        print("\nInfo: Scan data are being loaded successfully from:", file_name)

    def run_eye_scan(self, done: request.DoneFutureCallback = None) -> bool or request.CsFuture:
        """
        Kick off a run on 2D margin analysis and prepare scan data which is needed to draw an
        eye scan graph. The default scan settings will be used if users have not configured with
        new settings. See the following related commands for more info:

            save_eye_scan_data()
            load_eye_scan_data()
            display_eye_scan()

        Args:
            done: Optional command callback that will be invoked when the scan is finished

        Returns:
            Boolean value indicating success of the 2D margin scan

        """

        # Perform some logical data check on user settings
        prop_list = []
        prop_list.append("mgchk_rw_mode")
        prop_list.append("es_vref_min")
        prop_list.append("es_vref_max")
        prop_list.append("es_vref_steps")

        results = self.ddr_node.get_property(prop_list)
        rw_mode = results["mgchk_rw_mode"]
        vref_min = results["es_vref_min"]
        vref_max = results["es_vref_max"]
        vref_steps = results["es_vref_steps"]

        if rw_mode:
            if (vref_min > 50) or (vref_max > 50):
                print("ERROR: Cannot set Vref values larger than 50 under Write margin mode.")
                return False
            if vref_steps > 51:
                print("ERROR: Cannot set Vref step sizes larger than 51 for Write Margin mode")
                return False

        if vref_min > vref_max:
            print("ERROR: Cannot set minimum Vref value larger than maximum Vref value.")
            return False

        percent_min = self.get_eye_scan_vref_percentage(vref_min)
        percent_max = self.get_eye_scan_vref_percentage(vref_max)
        print("Min VRef is set at: ", percent_min, "%, value ", vref_min)
        print("Max VRef is set at: ", percent_max, "%, value ", vref_max)
        print("Number of VRef steps is set at: ", vref_steps)

        vref_incr_size = int((vref_max - vref_min) / (vref_steps - 1))
        max_step_size = vref_max - vref_min + 1

        if vref_incr_size < 1:
            print(
                "ERROR: For the selected VRef Min and VRef Max, the maximum step size is: ",
                str(max_step_size),
            )
            print(
                "Please either reduce the VRef steps or increase the VRef range by adjusting Min/Max."
            )
            return False

        scan_future = request.CsFutureSync(done)

        # Institute a Progress Bar
        progress_bar = PercentProgressBar()
        in_progress = "[bold dodger_blue2]In Progress[/]"
        progress_bar.add_task(
            description="2D Margin Scan Progress",
            status=PercentProgressBar.Status.STARTING,
            visible=True,
        )

        def progress_update(future):
            progress_bar.update(completed=future.progress * 100, status=in_progress)

        def done_scan(future):
            if future.error is not None:
                progress_bar.update(status=PercentProgressBar.Status.ABORTED)
            else:
                progress_bar.update(completed=100, status=PercentProgressBar.Status.DONE)

            if future.error:
                scan_future.set_exception(future.error)
            else:
                scan_future.set_result(None)

        def finalize(future):
            # Call on main thread for any final processing desired
            result_list = future._result
            if result_list[0]:
                # Can think about if want to provide auto-draw or
                # auto-save log for the users
                status = "COMPLETED"
            else:
                status = "ERROR"

            print("2D margin scan run status is: ", status)

        self.ddr_node.future(
            progress=progress_update, done=done_scan, final=finalize
        ).run_eye_scan_async(data_checked=True)
        print("Info: Margin Scans in progress...")

        return scan_future if done else scan_future.result

    def display_eye_scan(
        self, unit_index: int = 0, return_as_list: bool = False, display_type: str = "dynamic"
    ):
        """
        Assemble and display 2D eye scan drawing in chart format. By default it tries to process and
        draw from the scan data found in the most recent 2D eye scan run. Users have the option to
        load scan data first prior to using this function in order to display specific set of scan
        data, see load_eye_scan_data().

        By default, if no unit_index is specified, unit zero from the scan data will be displayed.
        For read margins, both rising and falling edge clock scan data will be processed and displayed.

        Args:
            unit_index: Specify the index from a set of scan data users intend to display, based on
                unit mode found in the scan settings (bit, nibble, byte)

            return_as_list: Optional argument, default to False. If set to True, the function will
                not display graphs by default. Instead, it returns a list of the grapghing objects
                back to the caller.

            display_type: Optional argument, default to dynamic. If set to static, a static image
                will be returned. If set to dynamic, a dynamic, interactive javascript view will be
                returned.

        Returns:
            A list of Figure object from Plotly, if return_as_list is specified as True,
            None is return otherwise by default.

        """
        if not _plotting_pkgs_available:
            raise ImportError(
                f"Plotting packages not installed! Please run ",
                "'pip install chipscopy[plotly,pandas]'",
            )

        result = self.ddr_node.get_property(["es_data_need_update"])
        data_need_update = result["es_data_need_update"]

        if data_need_update:
            self.eye_scan_data.clear()
            self.__update_eye_scan_data()

        if not self.eye_scan_data:
            print(
                "ERROR: No scan data is found. Please re-run 2D eye scan, "
                "or load scan data first from a data file."
            )
            return

        if isinstance(unit_index, str):
            print("ERROR: Please re-enter an integer value for unit_index.")
            return

        figure_list = []
        data_list = self.eye_scan_data
        scan_mode = data_list[0]["config.mode"]
        unit_mode = data_list[0]["config.unit"]
        rank_num = data_list[0]["config.rank"]
        mem_type = data_list[0]["config.mem_type"]
        df_vref = data_list[0]["config.df_vrefp"]
        df_vref = float(df_vref)
        nibble_count = data_list[0]["config.nibbles"]
        byte_count = data_list[0]["config.bytes"]
        unit_val = "{:02d}".format(unit_index)
        data_list = data_list[1:]

        # Some logical data check first on user inputs
        if unit_mode == "nibble":
            if (unit_index < 0) or (unit_index >= int(nibble_count)):
                max_unit = int(nibble_count) - 1
                err_msg = (
                    "ERROR: unit_index given for "
                    + unit_mode
                    + " mode needs to be between 0 and "
                    + str(max_unit)
                )
                err_msg += ", based on current memory configuration."
                print(err_msg)
                return
        elif unit_mode == "byte":
            if (unit_index < 0) or (unit_index >= int(byte_count)):
                max_unit = int(byte_count) - 1
                err_msg = (
                    "ERROR: unit_index given for "
                    + unit_mode
                    + " mode needs to be between 0 and "
                    + str(max_unit)
                )
                err_msg += ", based on current memory configuration."
                print(err_msg)
                return

        if "Read" in scan_mode:
            clocks = ["rise_", "fall_"]
            clock_defs = {
                "rise_": " - Rising Edge Clock - ",
                "fall_": " - Falling Edge Clock - ",
            }

            for clock in clocks:
                name_base = "read_" + clock + unit_val
                margin_mode = (
                    scan_mode
                    + clock_defs[clock]
                    + unit_mode.capitalize()
                    + str(unit_index)
                    + "<br>"
                )
                margin_mode += mem_type + " - Rank" + str(rank_num)
                fig = self.__draw_eye_scan_graph(data_list, name_base, margin_mode, df_vref)
                figure_list.append(fig)
        else:
            name_base = "write_" + unit_val
            margin_mode = scan_mode + " - " + unit_mode.capitalize() + str(unit_index) + "<br>"
            margin_mode += mem_type + " - Rank" + str(rank_num)
            fig = self.__draw_eye_scan_graph(data_list, name_base, margin_mode, df_vref)
            figure_list.append(fig)

        if not return_as_list:
            for fig in figure_list:
                if display_type == "dynamic":
                    fig.show()
                elif display_type == "static":
                    if not _jupyter_available:
                        raise ImportError(
                            f"Jupyter packages not installed! Please run 'pip install chipscopy[jupyter]'"
                        )
                    image_bytes = fig.to_image(format="png")
                    ipython_image = Image(image_bytes)
                    display(ipython_image)
                else:
                    raise ValueError("display_type argument must be static or dynamic")
        else:
            return figure_list

    def __str__(self):
        return self.name



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from typing import Tuple


@dataclass
class TraceInfo:
    data_width: int
    trigger_position: int
    # 'window_count' includes also any partial window.
    window_count: int
    window_size: int
    partial_window_sample_count: int = 0


RAW_DATA_SAMPLE_ALIGN_BIT_COUNT_32 = 32


def calculate_trace_info_captured(trace_info: TraceInfo, status_props: {}) -> TraceInfo:
    """
    Returns Trace_info, adjusting window_count and partial_window_sample_count
    window_count includes full and partial windows.
    """
    if status_props["is_full"]:
        return trace_info

    captured_window_count = status_props["windows_captured"]
    last_window_sample_count = status_props["samples_captured"] % trace_info.window_size
    if last_window_sample_count > trace_info.trigger_position:
        # Last window has trigger, keep the window
        captured_window_count += 1
    else:
        # Cannot keep partial window since it has not triggered.
        last_window_sample_count = 0

    return TraceInfo(
        data_width=trace_info.data_width,
        trigger_position=trace_info.trigger_position,
        window_count=captured_window_count,
        window_size=trace_info.window_size,
        partial_window_sample_count=last_window_sample_count,
    )


def calculate_raw_data_upload_byte_count(
    trace_info: TraceInfo, align_bit_count: int = RAW_DATA_SAMPLE_ALIGN_BIT_COUNT_32
) -> Tuple[int, int]:
    """
    Calculate total bytes, for all packets to upload.
    Have to upload full windows, even if a window is just partially filled.
    """
    # 1-bit trigger column, in raw data.
    align_byte_count = align_bit_count // 8
    one_sample_byte_count = align_byte_count * (
        (trace_info.data_width + 1 + align_bit_count - 1) // align_bit_count
    )
    wave_data_byte_count = one_sample_byte_count * trace_info.window_size * trace_info.window_count
    return one_sample_byte_count, wave_data_byte_count


def process_raw_data(
    trace_info: TraceInfo,
    raw_trace: bytearray,
    raw_data_align_bit_count=RAW_DATA_SAMPLE_ALIGN_BIT_COUNT_32,
) -> bytearray:
    """
    Order window samples, within samples.
    Remove trigger bit.
    Align samples on 1-byte boundary. (Raw data is aligned on 4-byte boundary)
    Remove invalid data from partially captured window.
    """
    if not raw_trace:
        raise ValueError("Uploaded waveform data is empty.")

    repacked_trace = align_samples_to_1_byte(
        raw_trace,
        trace_info.data_width + 1,
        trace_info.window_count * trace_info.window_size,
        raw_data_align_bit_count,
    )
    unrolled_wave = unroll_data(repacked_trace, trace_info)
    return unrolled_wave


def align_samples_to_1_byte(
    data: bytearray, sample_bit_count: int, sample_count: int, raw_data_sample_align_bit_count: int
) -> bytearray:
    """ Get rid of extra sample align bytes. Get rid of bytes beyond last sample."""
    in_sample_byte_count = calculate_data_sample_byte_count(
        raw_data_sample_align_bit_count, sample_bit_count
    )
    out_sample_byte_count = (sample_bit_count + 7) // 8
    valid_data_byte_count = min(len(data), in_sample_byte_count * sample_count)
    if in_sample_byte_count == out_sample_byte_count:
        return data[:valid_data_byte_count]

    # Remove most significant bytes, from each sample.
    move_len = out_sample_byte_count
    to_addr = 0

    for from_addr in range(0, valid_data_byte_count, in_sample_byte_count):
        data[to_addr : to_addr + move_len] = data[from_addr : from_addr + move_len]
        to_addr += move_len
    return data[:to_addr]


def calculate_data_sample_byte_count(
    data_sample_align_bit_count: int, valid_bit_count_per_sample: int
) -> int:
    """
    Args:
        data_sample_align_bit_count (int): Alignment bit count. Typically 32 bits (=4 bytes) for ILA.
        valid_bit_count_per_sample (int): Valid bits per sample, including any trigger bit.

    Returns: Number of bytes each sample occupies.
    """
    sample_word_count = (
        valid_bit_count_per_sample + data_sample_align_bit_count - 1
    ) // data_sample_align_bit_count
    sample_byte_count = sample_word_count * (data_sample_align_bit_count // 8)
    return sample_byte_count


def unroll_data(data: bytearray, trace_info: TraceInfo) -> bytearray:
    # Expect trigger column to be sample MSB, in in_data.
    # trace_info.data_width does not include trigger bit.
    # Samples aligned on byte boundary.
    # Returned data has no trigger column, which may make each sample one byte shorter.

    sample_byte_count_with_trigger = (trace_info.data_width + 1 + 7) // 8

    def unroll_one_window(window_data: memoryview) -> None:
        trigger_mark_sample_idx = -1
        first_trig_byte_addr, trig_bit_idx = divmod(trace_info.data_width, 8)
        trig_mask = 1 << trig_bit_idx
        for sample_idx, addr in enumerate(
            range(first_trig_byte_addr, len(window_data), sample_byte_count_with_trigger)
        ):
            if window_data[addr] & trig_mask != 0:
                trigger_mark_sample_idx = sample_idx
                break

        if trigger_mark_sample_idx < 0:
            raise ValueError("Corrupt waveform data. Missing window trigger mark.")

        start_of_window_sample = trigger_mark_sample_idx - trace_info.trigger_position
        if start_of_window_sample == 0:
            # Nothing to unroll.
            return

        first_sample_addr = sample_byte_count_with_trigger * start_of_window_sample
        if first_sample_addr < 0:
            # roll around
            first_sample_addr += len(window_data)

        buf = bytearray(len(window_data))
        buf[0 : len(window_data) - first_sample_addr] = window_data[first_sample_addr:]
        buf[len(window_data) - first_sample_addr :] = window_data[:first_sample_addr]
        window_data[:] = buf[:]

    def remove_trigger_byte(data: bytearray) -> bytearray:
        sample_byte_count_no_trigger = (trace_info.data_width + 7) // 8
        if sample_byte_count_no_trigger == sample_byte_count_with_trigger:
            return data

        # Remove most significant byte, from each sample.
        move_len = sample_byte_count_no_trigger
        to_addr = 0
        for from_addr in range(0, len(data), sample_byte_count_with_trigger):
            data[to_addr : to_addr + move_len] = data[from_addr : from_addr + move_len]
            to_addr += move_len
        return data[:to_addr]

    #
    window_byte_count = sample_byte_count_with_trigger * trace_info.window_size
    data_mv = memoryview(data)
    for window_addr in range(0, len(data), window_byte_count):
        # last window is not always full
        if window_addr + window_byte_count > len(data):
            window_byte_count = len(data) - window_addr
        unroll_one_window(data_mv[window_addr : window_addr + window_byte_count])

    res = remove_trigger_byte(data)
    return res



# *****************************************************************************
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

from .. import compat
from .AbstractChannel import AbstractChannel, EOS, EOM

ESC = 3


class StreamChannel(AbstractChannel):
    """Abstract channel implementation for stream oriented transport protocols.

    StreamChannel implements communication link connecting two end points
    (peers).
    The channel asynchronously transmits messages: commands, results and
    events.

    StreamChannel uses escape sequences to represent End-Of-Message and
    End-Of-Stream markers.

    Clients can subclass StreamChannel to support particular stream oriented
    transport (wire) protocol.
    Also, see ChannelTCP for a concrete IChannel implementation that works on
    top of TCP sockets as a transport.
    """

    def __init__(self, remote_peer, local_peer=None):
        super(StreamChannel, self).__init__(remote_peer, local_peer=local_peer)
        self.bin_data_size = 0
        self.buf = bytearray(0x1000)
        self.buf_pos = 0
        self.buf_len = 0

    def get(self):
        pass

    def put(self, n):
        pass

    def getBuf(self, buf):
        i = 0
        l = len(buf)
        while i < l:
            b = self.get()
            if b < 0:
                if i == 0:
                    return -1
                break
            buf[i] = b
            i += 1
            if i >= self.bin_data_size:
                break
        return i

    def putBuf(self, buf):
        for b in buf:
            self.put(b & 0xff)

    def read(self):
        while True:
            while self.buf_pos >= self.buf_len:
                self.buf_len = self.getBuf(self.buf)
                self.buf_pos = 0
                if self.buf_len < 0:
                    return EOS
            res = self.buf[self.buf_pos] & 0xff
            self.buf_pos += 1
            if self.bin_data_size > 0:
                self.bin_data_size -= 1
                return res
            if res != ESC:
                return res
            while self.buf_pos >= self.buf_len:
                self.buf_len = self.getBuf(self.buf)
                self.buf_pos = 0
                if self.buf_len < 0:
                    return EOS
            n = self.buf[self.buf_pos] & 0xff
            self.buf_pos += 1
            if n == 0:
                return ESC
            elif n == 1:
                return EOM
            elif n == 2:
                return EOS
            elif n == 3:
                for i in range(0, 100000, 7):
                    while self.buf_pos >= self.buf_len:
                        self.buf_len = self.getBuf(self.buf)
                        self.buf_pos = 0
                        if self.buf_len < 0:
                            return EOS
                    m = self.buf[self.buf_pos] & 0xff
                    self.buf_pos += 1
                    self.bin_data_size |= (m & 0x7f) << i
                    if (m & 0x80) == 0:
                        break
            else:
                if n < 0:
                    return EOS
                assert False

    def writeByte(self, n):
        if n == ESC:
            self.put(ESC)
            self.put(0)
        elif n == EOM:
            self.put(ESC)
            self.put(1)
        elif n == EOS:
            self.put(ESC)
            self.put(2)
        else:
            assert n >= 0 and n <= 0xff
            self.put(n)

    def write(self, buf):
        if isinstance(buf, int):
            self.writeByte(buf)
            return
        elif isinstance(buf, compat.strings):
            buf = bytearray(buf, 'utf-8')

        if len(buf) > 32 and self.isZeroCopySupported():
            self.put(ESC)
            self.put(3)
            n = len(buf)
            while True:
                if n <= 0x7f:
                    self.put(n)
                    break
                self.put((n & 0x7f) | 0x80)
                n >>= 7
            self.putBuf(buf)
        else:
            for b in buf:
                n = b & 0xff
                self.put(n)
                if n == ESC:
                    self.put(0)



# *****************************************************************************
# * Copyright (c) 2011, 2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import socket

from .. import compat
from .. import protocol
from .StreamChannel import StreamChannel


class ChannelTCP(StreamChannel):
    """ChannelTCP is a channel implementation that works on top of TCP sockets
    as a transport."""

    def __init__(self, remote_peer, host, port, client_socket = None):
        super(ChannelTCP, self).__init__(remote_peer)
        self.closed = False
        self.started = False
        channel = self

        class CreateSocket(object):
            def __call__(self):
                nonlocal host
                try:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

                    # NOTE - Useful for when on VPN.
                    if host.upper() == socket.gethostname().upper():
                        host = "localhost"

                    sock.connect((host, port))
                    sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                    channel.socket = sock
                    channel._onSocketConnected(None)
                except Exception as x:
                    channel._onSocketConnected(x)
        if not client_socket:
            protocol.invokeLater(CreateSocket())
        else:
            channel.socket = client_socket
            protocol.invokeLater(self._onSocketConnected, None)

    def _onSocketConnected(self, x):
        if x:
            self.terminate(x)
            self.closed = True
        if self.closed:
            try:
                if hasattr(self, 'socket') and self.socket:
                    self.socket.close()
            except socket.error as y:
                protocol.log("Cannot close socket", y)
        else:
            self.started = True
            self.start()

    def get(self):
        if self.closed:
            return -1
        try:
            return ord(self.socket.recv(1))
        except socket.error as x:
            if self.closed:
                return -1
            raise x

    def getBuf(self, buf):
        if self.closed:
            return -1
        try:
            return self.socket.recv_into(buf)
        except TypeError:
            # see http://bugs.python.org/issue7827
            # use super implementation
            self.getBuf = super(ChannelTCP, self).getBuf
            return self.getBuf(buf)
        except (ConnectionResetError, ConnectionAbortedError):
            protocol.invokeLater(self.close)
            return -1
        except socket.error as x:
            if self.closed:
                return -1
            raise x

    def str2bytes(self, data):
        if isinstance(data, compat.strings):
            return bytearray([ord(x) for x in data])
        elif isinstance(data, int):
            return bytearray([data])
        return data

    def put(self, b):
        if self.closed:
            return
        s = self.str2bytes(b)
        self.socket.send(s)

    def putBuf(self, buf):
        if self.closed:
            return
        if isinstance(buf, (bytes, bytearray)):
            s = buf
        else:
            s = self.str2bytes(buf)
        self.socket.sendall(s)

    def flush(self):
        pass

    def stop(self):
        self.closed = True
        if self.started:
            self.socket.close()



# %% [markdown]
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # ChipScoPy DDR Reporting Example
#
#
# <img src="../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
# This demo shows how to print and report DDR calibration status and report detailed information.
#
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Xilinx cs_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - Paths to server(s) and files are set correctly

# %%
import pprint
import os
import json
from chipscopy import create_session, report_versions
from chipscopy import get_design_files

# %%
# Specify locations of the running hw_server and cs_server below.
CS_URL = os.getenv("CS_SERVER_URL", "TCP:localhost:3042")
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")

# The get_design_files() function tries to find the PDI and LTX files. In non-standard
# configurations, you can put the path for PROGRAMMING_FILE and PROBES_FILE below.
design_files = get_design_files("vck190/production/chipscopy_ced")

PDI_FILE = design_files.programming_file
LTX_FILE = design_files.probes_file

print(f"HW_URL: {HW_URL}")
print(f"CS_URL: {CS_URL}")
print(f"PROGRAMMING_FILE: {PDI_FILE}")
print(f"PROBES_FILE:{LTX_FILE}")

# %% [markdown]
# ## 2 - Create a session and connect to the hw_server and cs_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout

# %%
# Start of the connection
session = create_session(cs_server_url=CS_URL, hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## 3 - Program the device with the example design
#
# After this step,
# - Device is programmed with the example programming file

# %%
# Typical case - one device on the board - get it.
versal_device = session.devices.filter_by(family="versal").get()
versal_device.program(PDI_FILE)

# %% [markdown]
# ## 4 - Discover Debug Cores
#
# Debug core discovery initializes the chipscope server debug cores. This brings debug cores in the chipscope server online.
#
# After this step,
#
# - The cs_server is initialized and ready for use

# %%
versal_device.discover_and_setup_cores()
print(f"Debug cores setup and ready for use.")

# %% [markdown]
# ## 5 - Show enabled DDRs in the device. Pick one to use

# %%
ddr_list = versal_device.ddrs
for ddr in ddr_list:
    print(ddr, "  Enabled:", ddr.is_enabled)

# Grab the first enabled DDR
for ddr in ddr_list:
    if ddr.is_enabled:
        print("Using Enabled DDR: ", ddr)
        break

# %% [markdown]
# ## 6 - Getting the Calibration Status
#
# There are several methods available to collect memory calibration information.

# %% [markdown]
# ### Method 1 - Calibration PASS/FAIL status

# %%
# Method 1 - Use Status string base API directly
print(ddr, "calibration status:", ddr.get_cal_status())

# %% [markdown]
# ### Method 2 - Calibration from the status property group

# %%
# Use Property Group to get dictionary base of results
props = ddr.ddr_node.get_property_group(["status"])
print(pprint.pformat(props, indent=2))

# %% [markdown]
# ### Method 3 - Detailed calibration status for each stage

# %%
# Use get Cal Stages API directly to also get dictionary results
props = ddr.get_cal_stages()
print(pprint.pformat(sorted(props.items()), indent=2))

# %% [markdown]
# ## 7 - Generate Full DDRMC Report
#
# The report() API call creates a full DDRMC status report to stdout or a file. This report includes memory configuration, margin analysis, calibration, and health status information.

# %%
# Use a single report command to get all latest essential
# Status and decoded data collected as it presents
ddr.report()
# Specify True to argument 1, and name/path to argument 2
# to get the report output generated and saved to a file
ddr.report(True, "test_out.txt")
print("Report Done.\n")

# %% [markdown]
# ## 8 - Dump the complete set of internal properties as json
#
# This demonstrates how to get a Python dictionary of all the low level DDR properties. These can be converted to JSON easily for export to other tools.

# %%
props = ddr.ddr_node.get_property_group([])
json_props = json.dumps(props, indent=4)
print(json_props)



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

"""
ZeroCopy dummy service to enable ZeroCopy binary transfers
"""

from ... import services
from ... import compat, protocol, services, channel, peer, errors


class ZeroCopyService(services.Service):
    instance = None

    def __init__(self):
        ZeroCopyService.instance = self

    def getName(self):
        return "ZeroCopy"


class ZeroCopyServiceProvider(services.ServiceProvider):
    def get_local_service(self, _channel):
        class CommandServer(channel.CommandServer):
            def command(self, token, name, data):
                pass
        _channel.addCommandServer(ZeroCopyService.instance, CommandServer())
        return (ZeroCopyService.instance,)

ZeroCopyService()
services.add_service_provider(ZeroCopyServiceProvider())


# *****************************************************************************
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""
Module protocol provides static methods to access Target Communication
Framework root objects:

1. the framework event queue and dispatch thread
2. local instance of Locator service, which maintains a list of available
   targets
3. list of open communication channels.

It also provides utility methods for posting asynchronous events,
including delayed events (timers).
"""

import sys
import threading
import time
import uuid

from . import EventQueue

_event_queue = None
_timer_dispatcher = None


def startEventQueue():
    global _event_queue, _timer_dispatcher
    if _event_queue and not _event_queue.isShutdown():
        return
    _event_queue = EventQueue.EventQueue(on_shutdown=shutdownDiscovery)
    _event_queue.start()
    # initialize LocatorService
    from .services.local.LocatorService import LocatorService
    from .services.local.ZeroCopyService import ZeroCopyService
    _event_queue.invokeLater(LocatorService)
    # start timer dispatcher
    _timer_dispatcher = threading.Thread(target=_dispatch_timers)
    _timer_dispatcher.setName("TCF Timer Dispatcher")
    _timer_dispatcher.setDaemon(True)
    _timer_dispatcher.start()


def stopEventQueue():
    """Stop the TCF event queue."""

    if not _event_queue or _event_queue.isShutdown():
        return

    _event_queue.shutdown()

    # Stop timer dispatcher

    global _timer_queue_alive

    try:
        with _timer_queue_lock:
            _timer_queue_alive = False
            _timer_queue_lock.notify_all()
    except Exception:
        pass


def getEventQueue():
    """
    @return instance of event queue that is used for TCF events.
    """
    return _event_queue


def isDispatchThread():
    """
    Returns true if the calling thread is the TCF event dispatch thread.
    Use this call to ensure that a given task is being executed (or not being)
    on dispatch thread.

    @return true if running on the dispatch thread.
    """
    return _event_queue is not None and _event_queue.isDispatchThread()


def invokeLater(c, *args, **kwargs):
    """
    Causes callable to be called with given arguments
    in the dispatch thread of the framework.
    Events are dispatched in same order as queued.
    If invokeLater is called from the dispatching thread
    the callable will still be deferred until
    all pending events have been processed.

    This method can be invoked from any thread.

    @param c - the callable to be executed asynchronously
    """
    _event_queue.invokeLater(c, *args, **kwargs)


def invokeLaterWithDelay(delay, c, *args, **kwargs):
    """
    Causes callable event to called in the dispatch thread of the framework.
    The event is dispatched after given delay.

    This method can be invoked from any thread.

    @param delay     milliseconds to delay event dispatch.
                     If delay <= 0 the event is posted into the
                     "ready" queue without delay.
    @param c - the callable to be executed asynchronously.
    """
    if delay <= 0:
        _event_queue.invokeLater(c, *args, **kwargs)
    else:
        with _timer_queue_lock:
            _timer_queue.append(Timer(time.time() + delay / 1000., c, *args,
                                      **kwargs))
            _timer_queue.sort()
            _timer_queue_lock.notify()


def invokeAndWait(c, *args, **kwargs):
    """
    Causes callable to be called in the dispatch thread of the framework.
    Calling thread is suspended until the method is executed.
    If invokeAndWait is called from the dispatching thread
    the callable is executed immediately.

    This method can be invoked from any thread.

    @param c  the callable to be executed on dispatch thread.
    """
    if _event_queue.isDispatchThread():
        return c(*args, **kwargs)
    else:
        class DoRun():
            result = None

            def __call__(self):
                try:
                    self.result = c(*args, **kwargs)
                finally:
                    with runLock:
                        runLock.notify()
        doRun = DoRun()
        runLock = threading.Condition()
        with runLock:
            _event_queue.invokeLater(doRun)
            runLock.wait()
            return doRun.result

_agentID = str(uuid.uuid4())


def getAgentID():
    return _agentID

_logger = None


def setLogger(logger):
    """
    Set framework logger.
    By default sys.stderr is used.

    @param logger - an object implementing Logger interface.
    """
    global _logger
    _logger = logger


def log(msg, x=None):
    """
    Logs the given message.
    @see #setLogger
    This method can be invoked from any thread.
    @param msg - log entry text
    @param x - an exception associated with the log entry or None.
    """
    if not _logger:
        sys.stderr.write(str(msg) + '\n')
        while x:
            import traceback
            sys.stderr.write("%s: %s\n" % (type(x).__name__, x))
            tb = getattr(x, "tb", None) or sys.exc_info()[2]
            if tb:
                traceback.print_tb(tb)
            caused_by = getattr(x, "caused_by", None)
            if caused_by:
                sys.stderr.write("Caused by: " + str(caused_by) + "\n")
                x = caused_by
            else:
                break
    else:
        _logger.log(msg, x)


def startDiscovery():
    "Start discovery of remote peers if not running yet"
    # initialize LocatorService
    from .services.local.LocatorService import LocatorService
    if LocatorService.locator:
        invokeAndWait(LocatorService.startup)


def shutdownDiscovery():
    "Shutdown discovery if running"
    from .services.local.LocatorService import LocatorService
    if LocatorService.locator:
        invokeAndWait(LocatorService.shutdown)


def getLocator():
    """
    Get instance of the framework locator service.
    The service can be used to discover available remote peers.
    @return instance of LocatorService.
    """
    from .services.local.LocatorService import LocatorService
    return LocatorService.locator


def getOpenChannels():
    """
    Return an array of all open channels.
    @return an array of IChannel
    """
    assert isDispatchThread()
    from . import transport
    return transport.getOpenChannels()


class ChannelOpenListener(object):
    """
    Interface to be implemented by clients willing to be notified when
    new TCF communication channel is opened.

    The interface allows a client to get pointers to channel objects that were
    opened by somebody else. If a client open a channel itself, it already has
    the pointer and does not need protocol.ChannelOpenListener. If a channel is
    created, for example, by remote peer connecting to the client, the only way
    to get the pointer is protocol.ChannelOpenListener.
    """
    def onChannelOpen(self, channel):
        pass


def addChannelOpenListener(listener):
    """
    Add a listener that will be notified when new channel is opened.
    @param listener
    """
    assert isDispatchThread()
    from . import transport
    transport.addChannelOpenListener(listener)


def removeChannelOpenListener(listener):
    """
    Remove channel opening listener.
    @param listener
    """
    assert isDispatchThread()
    from . import transport
    transport.removeChannelOpenListener(listener)


def sendEvent(service_name, event_name, data):
    """
    Transmit TCF event message.
    The message is sent to all open communication channels - broadcasted.
    """
    assert isDispatchThread()
    from . import transport
    transport.sendEvent(service_name, event_name, data)


def sync(done):
    """
    Call back after all TCF messages sent by this host up to this moment are
    delivered to their intended target. This method is intended for
    synchronization of messages across multiple channels.

    Note: Cross channel synchronization can reduce performance and throughput.
    Most clients don't need cross channel synchronization and should not call
    this method.

    @param done - will be executed by dispatch thread after pending
                  communication messages are delivered to corresponding
                  targets.
    """
    assert isDispatchThread()
    from . import transport
    transport.sync(done)


class CongestionMonitor(object):
    """
    Clients implement CongestionMonitor interface to monitor usage of local
    resources, like, for example, display queue size - if the queue becomes too
    big, UI response time can become too high, or it can crash all together
    because of OutOfMemory errors.

    TCF flow control logic prevents such conditions by throttling traffic
    coming from remote peers.

    Note: Local (in-bound traffic) congestion is detected by framework and
          reported to remote peer without client needed to be involved. Only
          clients willing to provide additional data about local congestion
          should implement CongestionMonitor and register it using
          protocol.addCongestionMonitor().
    """
    def getCongestionLevel(self):
        """
        Get current level of client resource utilization.
        @return integer value in range -100..100, where -100 means all
                resources are free, 0 means optimal load, and positive numbers
                indicate level of congestion.
        """
        raise NotImplementedError("Abstract method")

_congestion_monitors = []


def addCongestionMonitor(monitor):
    """
    Register a congestion monitor.
    @param monitor - client implementation of CongestionMonitor interface
    """
    assert isDispatchThread()
    _congestion_monitors.add(monitor)


def removeCongestionMonitor(monitor):
    """
    Unregister a congestion monitor.
    @param monitor - client implementation of CongestionMonitor interface
    """
    assert isDispatchThread()
    _congestion_monitors.remove(monitor)


def getCongestionLevel():
    """
    Get current level of local traffic congestion.

    @return integer value in range -100..100, where -100 means no pending
          messages (no traffic), 0 means optimal load, and positive numbers
          indicate level of congestion.
    """
    assert isDispatchThread()
    level = -100
    for m in _congestion_monitors:
        n = m.getCongestionLevel()
        if n > level:
            level = n
    if _event_queue:
        n = _event_queue.getCongestion()
        if n > level:
            level = n
    if level > 100:
        level = 100
    return level


def addServiceProvider(provider):
    """
    Register service provider.
    This method can be invoked from any thread.
    @param provider - ServiceProvider implementation
    """
    from . import services
    services.add_service_provider(provider)


def removeServiceProvider(provider):
    """
    Unregister service provider.
    This method can be invoked from any thread.
    @param provider - ServiceProvider implementation
    """
    from . import services
    services.removeServiceProvider(provider)


def addTransportProvider(provider):
    """
    Register transport provider.
    This method can be invoked from any thread.
    @param provider - TransportProvider implementation
    """
    from . import transport
    transport.addTransportProvider(provider)


def removeTransportProvider(provider):
    """
    Unregister transport provider.
    This method can be invoked from any thread.
    @param provider - TransportProvider implementation
    """
    from . import transport
    transport.removeTransportProvider(provider)


class Timer(object):
    timer_cnt = 0

    def __init__(self, time, run, *args, **kwargs):
        self.id = Timer.timer_cnt
        Timer.timer_cnt += 1
        self.time = time
        self.run = run
        self.args = args
        self.kwargs = kwargs

    def __lt__(self, x):
        return self.__cmp__(x) == -1

    def __cmp__(self, x):
        if x is self:
            return 0
        if self.time < x.time:
            return -1
        if self.time > x.time:
            return +1
        if self.id < x.id:
            return -1
        if self.id > x.id:
            return +1
        assert False
        return 0

_timer_queue_lock = threading.Condition()
_timer_queue = []
_timer_queue_alive = False


def _dispatch_timers():
    global _timer_queue_alive

    _timer_queue_alive = True

    try:
        with _timer_queue_lock:
            while _timer_queue_alive:
                if not _timer_queue:
                    _timer_queue_lock.wait()
                else:
                    tm = time.time()
                    t = _timer_queue[0]
                    if t.time > tm:
                        _timer_queue_lock.wait(t.time - tm)
                    else:
                        _timer_queue.pop(0)
                        invokeLater(t.run, *t.args, **t.kwargs)
    except RuntimeError:
        # Event queue is shut down, exit this thread
        pass
    except Exception as x:
        log("Exception in TCF dispatch loop", x)



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

from .. import memory
from ... import errors, channel
from ...channel.Command import Command


class Range(object):
    offs = 0
    size = 0
    stat = 0
    msg = None

    def __lt__(self, o):
        return self.__cmp__(o) == -1

    def __cmp__(self, o):
        if self.offs < o.offs:
            return -1
        if self.offs > o.offs:
            return +1
        return 0


class MemoryErrorReport(errors.ErrorReport, memory.MemoryError,
                        memory.ErrorOffset):
    def __init__(self, msg, attrs, addr, ranges):
        super(MemoryErrorReport, self).__init__(msg, attrs)
        if ranges is None:
            self.ranges = None
        else:
            self.ranges = []
            for m in ranges:
                r = Range()
                x = m.get(memory.ErrorOffset.RANGE_KEY_ADDR)
                if isinstance(x, str):
                    y = int(x)
                else:
                    y = x
                r.offs = y - addr
                r.size = m.get(memory.ErrorOffset.RANGE_KEY_SIZE)
                r.stat = m.get(memory.ErrorOffset.RANGE_KEY_STAT)
                key = memory.ErrorOffset.RANGE_KEY_MSG
                r.msg = errors.toErrorString(m.get(key))
                assert r.offs >= 0
                assert r.size >= 0
                self.ranges.append(r)
            self.ranges.sort()

    def getMessage(self, offset):
        if self.ranges is None:
            return None
        l = 0
        h = len(self.ranges) - 1
        while l <= h:
            n = int((l + h) / 2)
            r = self.ranges[n]
            if r.offs > offset:
                h = n - 1
            elif offset >= r.offs + r.size:
                l = n + 1
            else:
                return r.msg
        return None

    def getStatus(self, offset):
        if self.ranges is None:
            return memory.ErrorOffset.BYTE_UNKNOWN
        l = 0
        h = len(self.ranges) - 1
        while l <= h:
            n = int((l + h) / 2)
            r = self.ranges[n]
            if r.offs > offset:
                h = n - 1
            elif offset >= r.offs + r.size:
                l = n + 1
            else:
                return r.stat
        return memory.ErrorOffset.BYTE_UNKNOWN


class MemContext(memory.MemoryContext):
    def __init__(self, service, props):
        super(MemContext, self).__init__(props)
        self.service = service

    def fill(self, addr, word_size, value, size, mode, done):
        service = self.service
        contextID = self.getID()
        done = service._makeCallback(done)

        class FillCommand(MemoryCommand):
            def __init__(self):
                super(FillCommand, self).__init__(service, "fill",
                                                  (contextID, addr, word_size,
                                                   size, mode, value))

            def done(self, error, args):
                e = None
                if error:
                    # XXX : fle : Exception.message does not exist in python3,
                    #             better use str(Exception)
                    e = memory.MemoryError(str(error))
                else:
                    assert len(args) == 2
                    e = self.toMemoryError(addr, args[0], args[1])
                done.doneMemory(self.token, e)
        return FillCommand().token

    def get(self, addr, word_size, buf, offs, size, mode, done):
        service = self.service
        contextID = self.getID()
        done = service._makeCallback(done)

        class GetCommand(MemoryCommand):
            def __init__(self):
                super(GetCommand, self).__init__(service, "get",
                                                 (contextID, addr, word_size,
                                                  size, mode))

            def done(self, error, args):
                e = None
                if error:
                    # Exception.message does not exist in python3, better use
                    # str(Exception)
                    e = memory.MemoryError(str(error))
                else:
                    assert len(args) == 3
                    byts = channel.toByteArray(args[0])
                    assert len(byts) <= size
                    buf[offs:offs + len(byts)] = byts
                    e = self.toMemoryError(addr, args[1], args[2])
                done.doneMemory(self.token, e)
        return GetCommand().token

    def set(self, addr, word_size, buf, offs, size, mode, done):
        service = self.service
        contextID = self.getID()
        done = service._makeCallback(done)

        class SetCommand(MemoryCommand):
            def __init__(self):
                super(SetCommand, self).__init__(service, "set",
                                                 (contextID, addr, word_size,
                                                  size, mode,
                                                  bytearray(buf[offs:offs +
                                                                size])))

            def done(self, error, args):
                e = None
                if error:
                    # XXX : fle : Exception.message does not exist in python3,
                    #             better use str(Exception)
                    e = memory.MemoryError(str(error))
                else:
                    assert len(args) == 2
                    e = self.toMemoryError(addr, args[0], args[1])
                done.doneMemory(self.token, e)
        return SetCommand().token


class MemoryProxy(memory.MemoryService):
    def __init__(self, channel):
        super(MemoryProxy, self).__init__(channel)
        self.listeners = {}

    def getContext(self, context_id, done):
        done = self._makeCallback(done)
        service = self

        class GetContextCommand(Command):
            def __init__(self):
                super(GetContextCommand, self).__init__(service.channel,
                                                        service, "getContext",
                                                        (context_id,))

            def done(self, error, args):
                ctx = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    if args[1]:
                        ctx = MemContext(service, args[1])
                done.doneGetContext(self.token, error, ctx)
        return GetContextCommand().token

    def getChildren(self, parent_context_id, done):
        done = self._makeCallback(done)
        service = self

        class GetChildrenCommand(Command):
            def __init__(self):
                super(GetChildrenCommand, self).__init__(service.channel,
                                                         service,
                                                         "getChildren",
                                                         (parent_context_id,))

            def done(self, error, args):
                contexts = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    contexts = args[1]
                done.doneGetChildren(self.token, error, contexts)
        return GetChildrenCommand().token

    def addListener(self, listener):
        l = ChannelEventListener(self, listener)
        self.channel.addEventListener(self, l)
        self.listeners[listener] = l

    def removeListener(self, listener):
        l = self.listeners.get(listener)
        if l:
            del self.listeners[listener]
            self.channel.removeEventListener(self, l)


class MemoryCommand(Command):
    def __init__(self, service, cmd, args):
        super(MemoryCommand, self).__init__(service.channel, service, cmd,
                                            args)

    def toMemoryError(self, addr, data, ranges):
        if data is None:
            return None
        code = data.get(errors.ERROR_CODE)
        cmd = self.getCommandString()
        if len(cmd) > 72:
            cmd = cmd[0:72] + "..."
        msg = "TCF command exception:\nCommand: %s\n" % cmd + \
              "Exception: %s\nError code: %d" % (errors.toErrorString(data),
                                                 code)
        e = MemoryErrorReport(msg, data, addr, ranges)
        caused_by = data.get(errors.ERROR_CAUSED_BY)
        if caused_by is not None:
            e.caused_by = self.toError(caused_by, False)
        return e


class ChannelEventListener(channel.EventListener):
    def __init__(self, service, listener):
        self.service = service
        self.listener = listener

    def event(self, name, data):
        try:
            args = channel.fromJSONSequence(data)
            if name == "contextAdded":
                assert len(args) == 1
                self.listener.contextAdded(_toContextArray(self.service,
                                                           args[0]))
            elif name == "contextChanged":
                assert len(args) == 1
                self.listener.contextChanged(_toContextArray(self.service,
                                                             args[0]))
            elif name == "contextRemoved":
                assert len(args) == 1
                self.listener.contextRemoved(args[0])
            elif name == "memoryChanged":
                assert len(args) == 2
                self.listener.memoryChanged(args[0], _toAddrArray(args[1]),
                                            _toSizeArray(args[1]))
            else:
                raise IOError("Memory service: unknown event: " + name)
        except Exception as x:
            import sys
            x.tb = sys.exc_info()[2]
            self.service.channel.terminate(x)


def _toContextArray(svc, o):
    if o is None:
        return None
    ctx = []
    for m in o:
        ctx.append(MemContext(svc, m))
    return ctx


def _toSizeArray(o):
    if o is None:
        return None
    return [m.get("size", 0) for m in o]


def _toAddrArray(o):
    if o is None:
        return None
    return [m.get("addr") for m in o]



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from chipscopy.tcf import protocol
from chipscopy.tcf.channel import ChannelListener
from chipscopy.proxies.ChipScopeProxy import ChipScopeService, ChipScopeListener as CSListener
from chipscopy.utils.logger import log
from . import CsManager, Node, add_manager, get_manager, _managers, remove_manager

MANAGER_TYPE = "chipscope"

csm_pattern = re.compile(r"(.+)\." + MANAGER_TYPE)


def get_managers():
    for name, csm in _managers.items():
        if csm_pattern.match(name):
            yield csm


def add_manager_from_channel(channel):
    serv = channel.getRemoteService(ChipScopeService)
    if not serv:
        raise Exception("Server does not have ChipScope Service")

    # set up memory view manager
    name = "{}.{}".format(channel.remote_peer.getID(), MANAGER_TYPE)
    for manager in get_managers():
        if manager.name == name:
            if manager.channel != channel:
                remove_manager(manager)
                break
            return manager

    cs_manager = ChipScopeManager(name, channel)
    add_manager(cs_manager)

    # add channel listener for disconnection event
    channel.addChannelListener(ManagerChannelListener(cs_manager))

    # trigger update
    cs_manager.update_nodes()
    return cs_manager


def get_manager_from_channel(channel):
    channel_name = channel.remote_peer.getID()
    return get_manager(f"{channel_name}.{MANAGER_TYPE}")


def get_manager_from_channel_id(channel_id):
    return get_manager(f"{channel_id}.{MANAGER_TYPE}")


class ChipScopeListener(CSListener):
    def __init__(self, manager):
        self.manager = manager

    def node_added(self, node_ctx, props):
        log.dm.debug(f"{self.manager.name}: Adding Node {node_ctx}")
        parent_ctx = props.get("parent_ctx")
        if not parent_ctx:
            parent_ctx = ""
        node = self.manager.add_node(node_ctx, parent_ctx)
        node._props.update(props)  # avoid update notification
        self.manager._node_added()  # send notifications immediately

    def node_changed(self, node_ctx, props):
        log.dm.debug(f"{self.manager.name}: Changing Node {node_ctx}: {props}")
        try:
            node = self.manager[node_ctx]
            node.update(props)
            self.manager._node_updated(node)  # send notifications immediately
        except KeyError:
            # The presumption is that the node_ctx can't be found within the manager
            # this is the case when the server has cached state but the client doesn't know about a node for which
            # a node changed event just fired. In this specific case, we will create the node now
            parent_ctx = props.get("parent_ctx")
            if not parent_ctx:
                parent_ctx = ""
            node = self.manager.add_node(node_ctx, parent_ctx)
            node.update(props)  # avoid update notification
            self.manager._node_added()  # send notifications immediately

    def node_removed(self, node_ctx):
        log.dm.debug(f"{self.manager.name}: Removing Node {node_ctx}")
        try:
            node = self.manager[node_ctx]
            # node.invalidate()
            self.manager.remove_node(node.ctx)
            self.manager._node_removed()
        except KeyError:
            pass


class ChipScopeManager(CsManager):
    def __init__(self, name: str = "", channel=None):
        super(ChipScopeManager, self).__init__(name, channel)
        self.dc = None
        if self.channel:
            self.dc = self.channel.getRemoteService(ChipScopeService)
        if self.dc:
            self.dc.add_listener(ChipScopeListener(self))

    def update_nodes(self):
        self._update_children(self)

    def _update_children(self, parent):
        cs_manager = self

        def done_get_children(token, error, results):
            if not error:
                children = results
                if isinstance(children, str):
                    children = [children]
                cs_manager.set_children(parent, children)

                for ctx in children:
                    node = cs_manager.add_node(ctx, parent.ctx)
                    cs_manager._update_static_info(node)
                    cs_manager._update_children(node)
            cs_manager.remove_pending(token)

        if self.dc:
            self.add_pending(self.dc.get_children(parent.ctx, done_get_children))

    def _update_static_info(self, node: Node):
        def done_get_context(token, error, props):
            if not error:
                node.update(props)
            node.remove_pending(token)

        if self.dc:
            node.add_pending(self.dc.get_context(node.ctx, done_get_context))


class ChannelOpenListener(protocol.ChannelOpenListener):
    def onChannelOpen(self, channel):
        add_manager_from_channel(channel)


class ManagerChannelListener(ChannelListener):
    def __init__(self, cs_manager: ChipScopeManager):
        self.cs_manager = cs_manager

    def onChannelClosed(self, error):
        if self.cs_manager:
            self.cs_manager.invalidate()
            self.cs_manager.channel.removeChannelListener(self)


initialized = False


def init():
    global initialized
    if not initialized:
        if not protocol.getEventQueue():
            protocol.startEventQueue()
        protocol.invokeAndWait(protocol.addChannelOpenListener, ChannelOpenListener())
        initialized = True



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Any
from chipscopy.tcf.services import Service, DoneHWCommand, Token

JTAG_DEVICE_SERVICE = "JtagDevice"


class JtagDeviceProxy(Service):
    """TCF JtagCable service interface."""

    def __init__(self, channel):
        super(JtagDeviceProxy, self).__init__(channel)
        self.listeners = {}

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return JTAG_DEVICE_SERVICE

    def get_devices(self, done: DoneHWCommand = None) -> Token:
        """
        Get list of known idcodes
        :param done: Done callback
        :return: Token of request
        """
        return self.send_xicom_command("getDevices", (), done)

    def get_properties(self, idcode: int, done: DoneHWCommand = None) -> Token:
        """
        Get properties associated with idcode
        :param idcode: idcode of device
        :param done: Done callback
        :return: Token of request
        """
        return self.send_xicom_command("getProperties", (idcode,), done)

    def set_properties(self, props: Dict[str, Any], done: DoneHWCommand = None) -> Token:
        """
        Set properties associated with idcode
        :param props: Properties to set
        :param done: Done callback
        :return: Token of request
        """
        return self.send_xicom_command("setProperties", (props,), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum


class PerfMonTrafficQOSClass(enum.IntFlag):
    LOW_LATENCY_READ_QOS_CLASS = 0x1
    ISOCHRONOUS_READ_QOS_CLASS = 0x2
    BEST_EFFORT_READ_QOS_CLASS = 0x4
    ISOCHRONOUS_WRITE_QOS_CLASS = 0x8
    BEST_EFFORT_WRITE_QOS_CLASS = 0x10


TC_BEW = PerfMonTrafficQOSClass.BEST_EFFORT_WRITE_QOS_CLASS
TC_BER = PerfMonTrafficQOSClass.BEST_EFFORT_READ_QOS_CLASS
TC_ISOW = PerfMonTrafficQOSClass.ISOCHRONOUS_WRITE_QOS_CLASS
TC_ISOR = PerfMonTrafficQOSClass.ISOCHRONOUS_READ_QOS_CLASS
TC_LLR = PerfMonTrafficQOSClass.LOW_LATENCY_READ_QOS_CLASS



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict, Any
from chipscopy.tcf.channel import EventListener, fromJSONSequence
from chipscopy.tcf.services import Service, DoneHWCommand, Token
from chipscopy.tcf.services.arguments import from_xargs


NAME = "ChipScope"
"""ChipScope service name."""


class InvalidContextException(Exception):
    pass


class ChipScopeListener(object):
    """Debug node event listener is notified when debug node hierarchy changes,
    and when node is modified by DebugCore service commands.
    """

    def node_added(self, node_ctx: str, props: Dict[str, Any]):
        """
        Called when a new Debug Node was added
        :param node_ctx: Debug node context id
        :param props: Debug node properties added
        """
        pass

    def node_changed(self, node_ctx: str, props: Dict[str, Any]):
        """
        Called when a new Debug Node was changed
        :param node_ctx: Debug node context id
        :param props: Debug node properties changed
        """
        pass

    def node_removed(self, node_ctx: str):
        """
        Called when a new Debug Node was removed
        :param node_ctx: Debug node context id
        """
        pass


class ChipScopeProxy(Service):
    """TCF ChipScope service interface."""

    def __init__(self, channel):
        super(ChipScopeProxy, self).__init__(channel)
        self.listeners = {}

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def get_chipscope_server_version_info(self, done: DoneHWCommand) -> Token:
        """
        Returns the current package version
        :param done:
        :return: dict containing release specific version info
        """
        return self.send_xicom_command("getChipscopeServerVersionInfo", (), done)

    def get_children(self, parent_id: str, done: DoneHWCommand) -> Token:
        """
        Gets list of child nodes on a given hw_server channel and context id of parent node ("" is root).
        :param parent_id: Parent node ID
        :param done: Callback when command is complete
        :return: Token of request
        """
        return self.send_xicom_command("getChildren", (parent_id,), done)

    def get_context(self, node_id: str, done: DoneHWCommand) -> Token:
        """
        Gets context properties of a nodes on a given hw_server channel.
        :param node_id: Context node ID
        :param done: Callback when command is complete
        :return: Token of request
        """
        return self.send_xicom_command("getContext", (node_id,), done)

    def setup_debug_cores(
        self, node_id: str, debug_hub_addrs: List[int] = None, done: DoneHWCommand = None
    ) -> Token:
        """
        Sets up debug core environment from a given list of DebugHub
        :param node_id: Parent Context node ID
        :param debug_hub_addrs: List of DebugHub addresses
        :param done: Callback when command is complete
        :returns: Token of command request
        """
        props = {}
        if debug_hub_addrs:
            props["debug_hub_addrs"] = debug_hub_addrs
        return self.send_xicom_command("setupDebugCores", (node_id, props), done)

    def remove_debug_cores(self, node_id: str, done: DoneHWCommand = None) -> Token:
        """
        Removes debug nodes under a given parent debug node
        :param node_id: Parent Context node ID
        :param done: Callback when command is complete
        :returns: Token of command request
        """
        return self.send_xicom_command("removeDebugCores", (node_id,), done)

    def remote_connect_xvc(
        self,
        remote_id: str,
        xvc_host: str,
        xvc_port: str,
        idcode: int = 0,
        done: DoneHWCommand = None,
    ) -> Token:
        """
        Send a request to a remote server to connect to an xvc server at xvc_host:xvc_port
        :param remote_id: Remote server id or ChipScope manager id
        :param xvc_host: Hostname or ip address of XVC server
        :param xvc_port: Port of XVC server
        :param idcode: Optional parameter to force the XVC server to show a specific idcode
        :param done: Callback when command is complete
        :return: Token of request
        """
        return self.send_xicom_command(
            "remoteConnectXvc", (remote_id, xvc_host, xvc_port, idcode), done
        )

    def remote_disconnect_xvc(
        self, xvc_node_id: str, optional_args: Dict[str, Any] = None, done: DoneHWCommand = None
    ) -> Token:
        """
        Send a request to the remote server through which the XVC node is connected

        :param xvc_node_id: Context id of XVC node
        :param optional_args: Optional additional arguments
        +-------------------+----------------------+------------------------------------------+
        | Name              | Type                 | Description                              |
        +===================+======================+==========================================+
        | stop              | |bool|               | Stop the XVC prior to disconnect         |
        +-------------------+----------------------+------------------------------------------+
        :param done: Callback when command is complete
        :return: Token of request
        """
        if optional_args:
            args = (xvc_node_id, optional_args)
        else:
            args = (xvc_node_id,)
        return self.send_xicom_command("remoteDisconnectXvc", args, done)

    def get_css_param(self, param_names: List[str], done: DoneHWCommand = None) -> Token:
        return self.send_xicom_command("getCssParam", (param_names,), done)

    def set_css_param(self, param_names: Dict[str, Any], done: DoneHWCommand = None) -> Token:
        return self.send_xicom_command("setCssParam", (param_names,), done)

    def add_listener(self, listener: ChipScopeListener):
        """Add DebugCore service event listener.
        :param listener: Event listener implementation.
        """
        channel_listener = ChannelEventListener(self, listener)
        self.channel.addEventListener(self, channel_listener)
        self.listeners[listener] = channel_listener

    def remove_listener(self, listener: ChipScopeListener):
        """Remove DebugCore service event listener.
        :param listener: Event listener implementation.
        """
        channel_listener = self.listeners.get(listener)
        if channel_listener:
            del self.listeners[listener]
            self.channel.removeEventListener(self, channel_listener)


class ChannelEventListener(EventListener):
    def __init__(self, service, listener):
        self.service = service
        self.listener = listener
        self._event_handlers = {
            "nodesAdded": self._nodes_added,
            "nodeChanged": self._node_changed,
            "nodesRemoved": self._nodes_removed,
        }

    def _nodes_added(self, args):
        args = from_xargs(args)
        assert len(args) == 1
        nodes_props = args[0]
        for node_id, props in nodes_props.items():
            self.listener.node_added(node_id, props)

    def _node_changed(self, args):
        args = from_xargs(args)
        assert len(args) == 1
        for node_id, props in args[0].items():
            self.listener.node_changed(node_id, props)

    def _nodes_removed(self, args):
        args = from_xargs(args)
        assert len(args) == 1
        for node_id in args[0]:
            self.listener.node_removed(node_id)

    def event(self, name, data):
        try:
            args = fromJSONSequence(data)
            try:
                handler = self._event_handlers[name]
                handler(args)
            except KeyError:
                raise IOError("ChipScope service: unknown event: " + name)
        except Exception as x:
            import sys

            x.tb = sys.exc_info()[2]
            self.service.channel.terminate(x)


ChipScopeService = ChipScopeProxy



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chipscopy',
 'chipscopy._cli',
 'chipscopy.api',
 'chipscopy.api._detail',
 'chipscopy.api.ddr',
 'chipscopy.api.device',
 'chipscopy.api.ibert',
 'chipscopy.api.ibert.eye_scan',
 'chipscopy.api.ibert.link',
 'chipscopy.api.ibert.yk_scan',
 'chipscopy.api.ila',
 'chipscopy.api.noc',
 'chipscopy.client',
 'chipscopy.client.node',
 'chipscopy.client.util',
 'chipscopy.dm',
 'chipscopy.dm.harden',
 'chipscopy.dm.harden.noc_perfmon',
 'chipscopy.examples.ddr',
 'chipscopy.examples.ibert.versal_gty',
 'chipscopy.examples.ila_and_vio',
 'chipscopy.examples.memory',
 'chipscopy.examples.noc_perfmon',
 'chipscopy.examples.program',
 'chipscopy.examples.sysmon',
 'chipscopy.proxies',
 'chipscopy.shared',
 'chipscopy.tcf',
 'chipscopy.tcf.channel',
 'chipscopy.tcf.services',
 'chipscopy.tcf.services.local',
 'chipscopy.tcf.services.remote',
 'chipscopy.tcf.tests',
 'chipscopy.tcf.util',
 'chipscopy.utils',
 'chipscopy.utils.logger',
 'chipscopy.utils.noc_utils',
 'chipscopy.xvc']

package_data = \
{'': ['*'],
 'chipscopy': ['examples/designs/vck190/production/chipscopy_ced/*',
               'examples/designs/vmk180/production/chipscopy_ced/*',
               'examples/img/*'],
 'chipscopy.examples.ila_and_vio': ['img/*']}

install_requires = \
['Click>=7.1.2,<8.0.0',
 'importlib_metadata>=1.6.1,<2.0.0',
 'loguru>=0.5.0,<0.6.0',
 'more-itertools>=8.2.0,<9.0.0',
 'rich>=9.12.4,<10.0.0',
 'typing_extensions>=3.7.4,<4.0.0']

extras_require = \
{'core-addons': ['kaleido>=0.0.2,<0.0.3',
                 'plotly>=4.9.0,<5.0.0',
                 'pandas>=1.2.3,<2.0.0',
                 'matplotlib>=3.4.0,<4.0.0',
                 'PyQt5>=5.15.4,<6.0.0'],
 'jupyter': ['notebook>=6.0.3,<7.0.0', 'ipywidgets>=7.5.1,<8.0.0']}

entry_points = \
{'console_scripts': ['chipscopy = chipscopy._cli._chipscopy:main',
                     'chipscopy-get-examples = '
                     'chipscopy._cli.example_delivery:main']}

setup_kwargs = {
    'name': 'chipscopy',
    'version': '2021.2.1639266951',
    'description': 'Open-source project from Xilinx® that enables high-level control of Versal debug IP running in hardware',
    'long_description': '# 🐍 ChipScoPy README\n\n\n[![](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n\n![](https://raw.githubusercontent.com/Xilinx/chipscopy/master/docs/images/chipscopy_logo_head_right_transparent_background.png)\n\nChipScoPy is an open-source project from Xilinx® that enables high-level control of Versal debug IP running in hardware.\nUsing a simple Python API, developers can control and communicate with ChipScope® debug IP such as the Integrated Logic\nAnalyzer (ILA), Virtual IO (VIO), device memory access, and more.\n\n-------------------------------------------------------------------------------\n\n![](https://raw.githubusercontent.com/Xilinx/chipscopy/master/docs/images/chipscopy_overview.png)\n\n-------------------------------------------------------------------------------\n\n[ChipScoPy Overview](https://xilinx.github.io/chipscopy/2021.1/overview.html)\n\n-------------------------------------------------------------------------------\n\n[System Requirements](https://xilinx.github.io/chipscopy/2021.1/system_requirements.html)\n\n-------------------------------------------------------------------------------\n\n[ChipScoPy Installation](https://xilinx.github.io/chipscopy/2021.1/chipscopy_installation.html)\n\n-------------------------------------------------------------------------------\n\n[ChipScoPy Examples](https://github.com/Xilinx/chipscopy/tree/master/chipscopy/examples)\n\n-------------------------------------------------------------------------------\n\n[API Documentation](https://xilinx.github.io/chipscopy/)\n\n-------------------------------------------------------------------------------\n',
    'author': 'Xilinx ChipScope Team',
    'author_email': None,
    'maintainer': 'Xilinx ChipScope Team',
    'maintainer_email': None,
    'url': 'https://github.com/Xilinx/chipscopy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

from .. import runcontrol
from ... import channel
from ...channel.Command import Command


class RunContext(runcontrol.RunControlContext):
    def __init__(self, service, props):
        super(RunContext, self).__init__(props)
        self.service = service

    def getISA(self, address, done):
        service = self.service
        done = service._makeCallback(done)
        contextID = self.getID()

        class GetISACommand(Command):
            def __init__(self):
                super(GetISACommand, self).__init__(service.channel, service,
                                                    "getISA",
                                                    (contextID, address))

            def done(self, error, args):
                isa = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    isa = runcontrol.RunControlISA(args[1])
                done.doneGetISA(self.token, error, isa)
        return GetISACommand().token

    def getState(self, done):
        service = self.service
        done = service._makeCallback(done)
        contextID = self.getID()

        class GetStateCommand(Command):
            def __init__(self):
                super(GetStateCommand, self).__init__(service.channel, service,
                                                      "getState", (contextID,))

            def done(self, error, args):
                susp = False
                pc = None
                reason = None
                states = None
                if not error:
                    assert len(args) == 5
                    error = self.toError(args[0])
                    susp = args[1]
                    if args[2] is not None:
                        pc = int(args[2])
                    reason = args[3]
                    states = args[4]
                done.doneGetState(self.token, error, susp, pc, reason, states)
        return GetStateCommand().token

    def detach(self, done):
        return self._command("detach", (self.getID(),), done)

    def resume(self, mode, count, params, done):
        if not params:
            return self._command("resume", (self.getID(), mode, count), done)
        else:
            return self._command("resume", (self.getID(), mode, count, params),
                                 done)

    def suspend(self, done):
        return self._command("suspend", (self.getID(),), done)

    def terminate(self, done):
        return self._command("terminate", (self.getID(),), done)

    def _command(self, cmd, args, done):
        service = self.service
        done = service._makeCallback(done)

        class RCCommand(Command):
            def __init__(self, cmd, args):
                super(RCCommand, self).__init__(service.channel, service, cmd,
                                                args)

            def done(self, error, args):
                if not error:
                    assert len(args) == 1
                    error = self.toError(args[0])
                done.doneCommand(self.token, error)
        return RCCommand(cmd, args).token


class ChannelEventListener(channel.EventListener):
    def __init__(self, service, listener):
        self.service = service
        self.listener = listener

    def event(self, name, data):
        try:
            args = channel.fromJSONSequence(data)
            if name == "contextSuspended":
                assert len(args) == 4
                self.listener.contextSuspended(args[0], args[1], args[2],
                                               args[3])
            elif name == "contextResumed":
                assert len(args) == 1
                self.listener.contextResumed(args[0])
            elif name == "contextAdded":
                assert len(args) == 1
                self.listener.contextAdded(_toContextArray(self.service,
                                                           args[0]))
            elif name == "contextChanged":
                assert len(args) == 1
                self.listener.contextChanged(_toContextArray(self.service,
                                                             args[0]))
            elif name == "contextRemoved":
                assert len(args) == 1
                self.listener.contextRemoved(args[0])
            elif name == "contextException":
                assert len(args) == 2
                self.listener.contextException(args[0], args[1])
            elif name == "containerSuspended":
                assert len(args) == 5
                self.listener.containerSuspended(args[0], args[1], args[2],
                                                 args[3], args[4])
            elif name == "containerResumed":
                assert len(args) == 1
                self.listener.containerResumed(args[0])
            elif name == "contextStateChanged":
                assert len(args) == 1
                self.listener.contextStateChanged(args[0])
            else:
                raise IOError("RunControl service: unknown event: " + name)
        except Exception as x:
            import sys
            x.tb = sys.exc_info()[2]
            self.service.channel.terminate(x)


class RunControlProxy(runcontrol.RunControlService):
    def __init__(self, channel):
        self.channel = channel
        self.listeners = {}

    def addListener(self, listener):
        l = ChannelEventListener(self, listener)
        self.channel.addEventListener(self, l)
        self.listeners[listener] = l

    def removeListener(self, listener):
        l = self.listeners.get(listener)
        if l:
            del self.listeners[listener]
            self.channel.removeEventListener(self, l)

    def getContext(self, context_id, done):
        done = self._makeCallback(done)
        service = self

        class GetContextCommand(Command):
            def __init__(self):
                super(GetContextCommand, self).__init__(service.channel,
                                                        service, "getContext",
                                                        (context_id,))

            def done(self, error, args):
                ctx = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    if args[1]:
                        ctx = RunContext(service, args[1])
                done.doneGetContext(self.token, error, ctx)
        return GetContextCommand().token

    def getChildren(self, parent_context_id, done):
        done = self._makeCallback(done)
        service = self

        class GetChildrenCommand(Command):
            def __init__(self):
                super(GetChildrenCommand, self).__init__(service.channel,
                                                         service,
                                                         "getChildren",
                                                         (parent_context_id,))

            def done(self, error, args):
                contexts = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    contexts = args[1]
                done.doneGetChildren(self.token, error, contexts)
        return GetChildrenCommand().token


def _toContextArray(svc, o):
    if o is None:
        return None
    ctx = []
    for m in o:
        ctx.append(RunContext(svc, m))
    return ctx
